/*
 * bitonic.sc - Split-C sample bitonic sort program
 *
 * "Copyright (c) 1993 by The Regents of the University of California.
 * All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Mike Dahlin
 * Version:			174
 * Creation Date:	Fri Jan 22 12:11:57 1993
 * Filename:		bitonic.sc
 * History:
 *	Rich Martin	1	Tue Jun  1 23:04:26 PDT 1993
 *	Prepared for release as example.
 */
			
/******************** INCLUDES **************************/

#include <stdio.h>
#include <split-c/split-c.h>
#include <split-c/bulk.h>
#include <split-c/com.h>
#include "modes.h"
#include "macros.h"
#include "bitonic.h"

int data[PROCS]::[MAX_PER_PROC];
int temp[PROCS]::[MAX_PER_PROC];
#ifndef MEMORY
int *localReturn;
#else
int localReturn[MAX_PER_PROC];
#endif

double sort_time, sstart, sfinish; /* time for local sorts */
double cmp_time, cmp_start;
double merge_time, merge_start;
double xfer_time, xfer_start;

int logPROCS;
int *localData;
int *localTemp;
int k = (1024*64);    /* number of elements per processor */

void splitc_main(int argc, char **argv) 
{

  int check = 0;
  int seed = 17;
  int merge = 0;
  int i;
  double start,finish;

  /* parse command line args */
  argv++;
  argc--;
  while (argc>0) {
    if (**argv=='-')
      switch (*((*argv)+1)) {
      case 'h':
	on_one {
	  printf("Bitonic sort options: (defaults)\n");
	  printf("  -n number of keys (in K) per processor  (64K)\n");
	  printf("  -s seed value\n");
	  printf("  -m merge algorithm\n");
	  printf("\t<0 -- each proc. calc entire block independently>\n");
	  printf("\t<1 -- pairs of procs calc 1/2 of block and swap results>\n");
	  printf("  -c check result\n");
	}
	break;
      case 'n':
	if (--argc) sscanf(*++argv, "%d", &k);
	k *= 1024;
	if (k > MAX_PER_PROC) {
	  on_one {
	    printf("ERROR: k (%d) can not exceed %d\n", k, MAX_PER_PROC);
	    exit(-1);
	  }
	}
	break;
      case 's':
	if (--argc) sscanf(*++argv, "%d", &seed);
	break;
      case 'm':
	if (--argc) sscanf(*++argv, "%d", &merge);
	break;
      case 'c':
	check = 1;
	break;
      }
    else printf("Argument %s meaningless.\n",*argv);
    argc--;
    argv++;

  }
  on_one  {
    printf ("Bitonic sort of %d keys on %d processors\n", k, PROCS);
    switch (merge) {
    case 0:
      printf("Using full store merge algorithm.\n");
      break;
    case 1:
      printf("Using half store  merge algorithm.\n");
      break;
    }
    printf("Seed: %d\n", seed);

  }


  /* compute log_2 of the number of processors.  Make sure it's integral */
  for (i = 1, logPROCS = 0; i < PROCS; i = i << 1, logPROCS++);
  if (i != PROCS) {
    printf("Couldn't calculate log of PROCS.  Exiting.\n");
    exit(1);
  }
  /* 
   * Seed random generator based on processor number
   * so all processors dont have same collection of numbers 
   */
  srandom( (unsigned int) MYPROC + seed);
  
  /* allocate space */
  localData = (int *) data;
  localTemp = (int *) temp; 
#ifndef MEMORY
  localReturn = (int *)temp + (k>>1);
#endif
  
  /* initialize local values to random */
  for(i = 0; i < k; i++){
    localData[i] = random();
  }


  sort_time = cmp_time = merge_time = xfer_time = 0.0;

  /***   Set up timers:   ***/
  barrier();  
  start=get_seconds();
  
  sstart = get_seconds();
  local_radix_sort(0, k,localData, localTemp);
  sfinish = get_seconds();
  sort_time += sfinish-sstart;
  barrier();
  bitonic_sort(merge);

  finish=get_seconds();
  
  barrier();
  /* print results */
#if 0
  on_one {
    printf("\n++++ Timing Results ++++\n");
    printf("xfer: \t%.3lf\n", xfer_time);
    printf("cmp: \t%.3lf \n", cmp_time);
    printf("merge: \t%.3lf \n", merge_time);
    printf("\nlocal sort: %.3lf\n",sort_time);
    printf ("\nTotal Time: %.3lf \n\nKeys Sorted per Proc per Sec: %.3lf\n\n",
	    finish - start, k/(finish - start));
  }

#else
  on_one printf("Procs: %d  Keys/PROCS: %d  Total -- %.3lf   merge -- %.3lf   sort -- %.3lf\n",
		PROCS, k, finish - start, merge_time, sort_time);
#endif

  barrier();
  if (check) checkData(k);

}


void   bitonic_sort(int merge)
{
  int outerdim, innerdim;
  int compMode, mode;
  int ii;

  /* MAIN LOOP */
  /* major and minor step correspond to the number of processors 
     in a group working on the same merge */
  
  /* We do logPROCS major steps.
     at each step, the processors in an outerdim-dimension subcube
     are working together to do the bitonic merge on their
     collection of the inputs.  See the paper on CM-2 sorting for details */
  
  for(outerdim = 1; outerdim <= logPROCS; outerdim ++){
    
    /* for each major step, do a cube swap for dimensions within
       this group of processors.  */
    for(innerdim = outerdim - 1; innerdim >= -1; innerdim --){

      /*  BASE CASE */
      if(innerdim == -1) {
	/* At the base case, we would be doing "cube swaps" in
	   virtual dimensions which reside completely within a single
	   physical processor.  In this case, just do a local sort
	   instead. */
	
	sstart = get_seconds();
	local_radix_sort(outerdim, k,localData,localTemp);  		
	sfinish=get_seconds();
	sort_time += sfinish-sstart;

      } 
      else {
	merge_start = get_seconds();
	bitonic_merge(merge, innerdim, outerdim);
	merge_time += get_seconds() - merge_start;
      }
    }
  }
}

void   bitonic_merge(int merge, int i, int o)
{

  if (merge == 0)
    bitonic_merge_full(i, o);
  else
    bitonic_merge_half(i, o);

}

void  bitonic_merge_half(int innerdim, int outerdim)
{
  int exchangeWith;	/* exchangeWith is who I will
			   send my localData to and receive localTemp
			   from when we do the communication step */
  int compMode;
  int i, half_keys;
#ifndef MEMORY
  int j;
#endif
  
  /* actually do the cube swap in a physical dimension
     with a neighbor */
  xfer_start = get_seconds();

  half_keys = k>>1;
  compMode = getMode( MYPROC, innerdim, outerdim);
  exchangeWith = getBuddy(MYPROC, innerdim);

  if (compMode == COMP_MIN)
#ifndef MEMORY
    /* Give the processor calculating the max your top half of data */
    bulk_store(temp[exchangeWith], localData + half_keys, 
	       half_keys*sizeof(int));
#else
    bulk_store(temp[exchangeWith] + half_keys, localData + half_keys, 
	       half_keys*sizeof(int));
#endif
  else
    /* Give the processor calculating the min your bottom half of data */
    bulk_store(temp[exchangeWith], localData, half_keys*sizeof(int));

  all_store_sync();
  xfer_time += get_seconds() - xfer_start;

  /* after a cube swap, one partner keeps the MIN, the other takes the MAX */
  cmp_start = get_seconds();

  if (compMode == COMP_MIN)  {
    /* Look only at the bottom half of the data */
    for (i = 0; i < half_keys; i++)   
      if (localData[i] > localTemp[i]) {
	localReturn[i] = localData[i];
	localData[i] = localTemp[i];
      }
      else
	localReturn[i] = localTemp[i];
    bulk_store(data[exchangeWith], localReturn, half_keys*sizeof(int));
  }
  else  {
#ifndef MEMORY
    int *Dptr, *Tptr, *Rptr;

    Dptr = localData + half_keys;
    Tptr = localTemp;
    Rptr = localReturn;
    for(i = 0; i < half_keys ; i++, Dptr++) 
      if (*Tptr < *Dptr) {
	*Rptr++ = *Dptr;
	*Dptr = *Tptr++;
      }
      else
	*Rptr++ = *Tptr++;
    bulk_store(data[exchangeWith]+half_keys, localReturn, half_keys*sizeof(int));
#else
    for(i = half_keys; i < k; i++) 
      if (localData[i] < localTemp[i]) {
	localReturn[i] = localData[i];
	localData[i] = localTemp[i];
      }
      else
	localReturn[i] = localTemp[i];
    bulk_store(data[exchangeWith]+half_keys, localReturn+half_keys, half_keys*sizeof(int));
#endif

  }
  all_store_sync();
  cmp_time += get_seconds() - cmp_start;
}  

void   bitonic_merge_full(int innerdim, int outerdim)
{
  int exchangeWith;	/* exchangeWith is who I will
			   send my localData to and receive localTemp
			   from when we do the communication step */
  int compMode;
  int i;

  /* actually do the cube swap in a physical dimension
     with a neighbor */
  xfer_start = get_seconds();
  exchangeWith = getBuddy(MYPROC, innerdim);

  bulk_store(toglobal(exchangeWith,localTemp), localData, k*sizeof(int));
  all_store_sync();


  xfer_time += get_seconds() - xfer_start;

  /* after a cube swap, one partner keeps the MIN, the other takes the MAX */
  cmp_start = get_seconds();
  compMode = getMode( MYPROC, innerdim, outerdim);

  if (compMode == COMP_MIN) {
    for (i = 0; i < k; i++)   
      if (localData[i]>localTemp[i]) 
	localData[i] = localTemp[i];
  }
  else  {
    for(i = 0; i < k; i++) 
      if ( localData[i] < localTemp[i]) 
	localData[i] =localTemp[i];
  }

  cmp_time += get_seconds() - cmp_start;

}

#if 0
/* returns who processor iproc should exchange data with when the inner 
   loop is working on dimension innderdim */
int   getBuddy(int iproc, int innerdim)
{
  if(iproc & (BITMASK(innerdim)))
    return(iproc - (OFFSET(innerdim)));

  return(iproc + (OFFSET(innerdim)));
}


int   getMode(int iproc, int innerdim, int outerdim)
{
  /* xor of bit (innerdim) of iproc and bit outerdim of iproc */
  int innerbit, outerbit;

  innerbit = iproc & (BITMASK(innerdim));
  outerbit = iproc & (BITMASK(outerdim));

  if((innerbit || outerbit) && !(innerbit && outerbit))
    return COMP_MIN;
    
  return COMP_MAX;
}

static int check[PROCS]::;

void checkData(int k)
{
  int ii,jj;
  int error = 0;
  on_one printf("Checking correctness of sort. . .\n");

  for(ii = 0; ii < (k-1); ii++){
    if (data[MYPROC][ii] < data[MYPROC][ii+1]) {
      printf("Error at proc %d item %d\n",MYPROC,ii);
      error++;
    }
  }
  check[MYPROC]=data[MYPROC][k-1];
  barrier();
  if (MYPROC < (PROCS - 1)) {
    if (check[MYPROC] < check[MYPROC+1]) {
      printf("proc %d: error across to  %d \n",MYPROC,MYPROC+1);  
      error++;
    }
  }
  if (!error) on_one printf(". . .keys sorted correctly.\n");
}

#endif
