/*									tab:4
 *
 * cannon.sc - global matrix multiplication example usng Cannons
 * algorithm.
 *
 * This example is intended to demonstrate how systolic algorithms 
 * may be implemented in split-c.  These typically involve alternating
 * phases of local computation and phases of communication along a
 * fixed regular interconnection.  In the case of Cannon's algorithm,
 * the local computation is a matrix multiplication of submatrixes
 * and the communication is shifts of the submatrixes along rows and
 * columns of the outer grid.
 *
 * The local matrix multiply is highly optimized and the shifts are
 * accomplished with bulk_stores and all_store_sync.
 *
 * The example includes timing of the global multiply and the
 * two individual phases.
 *
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Arvind Krishnamurthy and David E. Culler
 * Version:			8
 * Creation Date:	Thu Feb  4 22:41:14 1993
 * Filename:		cannon.sc
 * History:
 *	ssl	7	Fri Dec 10 00:50:37 1993
 *		Use pow() instead of exp2(). Paragon doesn't have exp2().
 *	ssl	6	Mon Oct 25 23:29:03 1993
 *		Fixed scanf. Stdin always opens in independent mode.
 *	LTL	5	Mon Aug 16 17:25:37 1993
 *		Added code to use scanf instead of getInt for CMMD version
 *	AK	3	Wed Jul  7 01:11:38 1993
 *		Replace the gswap function by a macro. Blame C and not Split-C!
 * 	TvE	2	Sun Apr  4 00:18:56 PST 1993
 *		changed all_bcast to all_bcast_i
 *	DEC	1	Fri Feb  5 01:07:15 1993
 *		Initial version
 *
 */

static char _version_string_[] = "\nVersion:7:cannon.sc\0\n";

#include <stdio.h>
#include <math.h>

#include <split-c/split-c.h>
#include <split-c/control.h>
#include <split-c/com.h>

#define gswap(A, B) { void *global T = (*A); (*A) = (*B); (*B) = T; }

void all_canon_mm(int l, int n, int r, int m, 
				  double A[l][l]::[n][r],
				  double B[l][l]::[r][m],
				  double C[l][l]::[n][m])
{
	/* This is just the way you declare in C. Only difference
	   is we have a "spread" inbetween the "*" and the var name.
	*/
    double (*spread nA)[l]::[n][r];
    double (*spread nB)[l]::[r][m];

	/* Example use of all_spread_malloc. Brings out the
	   2D nature of the address space. You specify the number
	   of elements to be spread across the procs, as well as
	   the number of units in each processor.
	*/
    nA = all_spread_malloc (l*l, n*r*sizeof(double));
    nB = all_spread_malloc (l*l, r*m*sizeof(double));

    {
      int numSteps;
      int x,y;
      int i,j,tmp;

      /* Initialize result matrix C */
      for_my_2D(i,j,tmp,l,l)	
		  zero_matrix(n,m,tolocal(C[i][j]));

      /* Skew input matrixes */
      for_my_2D(i,j,tmp,l,l) {              
		  bulk_store(nA[i][(j-i+l)%l], tolocal(A[i][j]), n*r*sizeof(double));
		  bulk_store(nB[(i-j+l)%l][j], tolocal(B[i][j]), r*m*sizeof(double));
      }
      gswap(&nA,&A);
      gswap(&nB,&B);
      all_store_sync();

      /* Multiply and shift */

      for (numSteps=0; numSteps<l-1 ; numSteps++) {
		  for_my_2D(i,j,tmp,l,l) {              
			  matrix_mult(n, r, m, tolocal(C[i][j]), 
						  tolocal(A[i][j]), tolocal(B[i][j]));
			  bulk_store(nA[i][(j-1+l)%l], tolocal(A[i][j]), 
						 n*r*sizeof(double));
			  bulk_store(nB[(i-1+l)%l][j], tolocal(B[i][j]), 
						 r*m*sizeof(double));
		  }
		  gswap(&nA,&A);
		  gswap(&nB,&B);
		  all_store_sync();

      }

      /* Final Multiply */
      for_my_2D(i,j,tmp,l,l) {              
		  matrix_mult(n, r, m, tolocal(C[i][j]),
					  tolocal(A[i][j]), tolocal(B[i][j]));
      }
      barrier();
    }
}

/* Test out the data movement with no local computation */
void all_cannon_move(int l, int n, int r, int m, 
	      double A[l][l]::[n][r],
	      double B[l][l]::[r][m],
	      double C[l][l]::[n][m])
{

    double (*spread nA)[l]::[n][r];
    double (*spread nB)[l]::[r][m];

    nA = all_spread_malloc (l*l, n*r*sizeof(double));
    nB = all_spread_malloc (l*l, r*m*sizeof(double));

    {
      int numSteps;
      int x,y;
      int i,j,tmp;

      /* Skew input matrixes */

      for_my_2D(i,j,tmp,l,l) {              
		  bulk_store(nA[i][(j-i+l)%l], tolocal(A[i][j]), n*r*sizeof(double));
		  bulk_store(nB[(i-j+l)%l][j], tolocal(B[i][j]), r*m*sizeof(double));
      }
      gswap(&nA,&A);
      gswap(&nB,&B);
      all_store_sync();

      /* Shift */
      for (numSteps=0; numSteps<l-1 ; numSteps++) {
		  for_my_2D(i,j,tmp,l,l) {              
			  bulk_store(nA[i][(j-1+l)%l], tolocal(A[i][j]), 
						 n*r*sizeof(double));
			  bulk_store(nB[(i-1+l)%l][j], tolocal(B[i][j]),
						 r*m*sizeof(double));
		  }
		  gswap(&nA,&A);
		  gswap(&nB,&B);
		  all_store_sync();
      }
    }
}


/* Perform local multiplies on submatrices, without any communication to 
   determine peak multiply performance */

void all_fake_mm(int l, int n, int r, int m, 
				 double A[l][l]::[n][r],
				 double B[l][l]::[r][m],
				 double C[l][l]::[n][m])
{
  int numSteps;
  int x,y;
  int i,j,tmp;

  /* Initialize result matrix C */
  for_my_2D(i,j,tmp,l,l)	
    zero_matrix(n,m,tolocal(C[i][j]));

  for (numSteps=0; numSteps<l ; numSteps++) {
	  barrier();
	  for_my_2D(i,j,tmp,l,l) {              
		  matrix_mult(n, r, m, tolocal(C[i][j]),
					  tolocal(A[i][j]), tolocal(B[i][j]));
	  }
  }
  barrier();
}


void splitc_main() {
    int l,n,m,r;
    int L,N,M,R;
    double secs;

/* By default, stdin is independent mode and
 * stdout is in global independent mode.
 */

    on_one {
		printf("Input Size of Matrix:\n");
		printf("Square grid of side length L and small patches of "
			   "(N*R), (R*M), and (N*M)\n");
		printf("Matrix Multiply: (L*N)x(L*R) X (L*R)x(L*M) => (L*N)x(L*M)\n");
		printf("Enter L N R M:");
		fflush(stdout);
    }
	on_one
		scanf("%d", &l);
	l = all_bcast_i(l);
	on_one
		scanf("%d", &n);
	n = all_bcast_i(n);
	on_one
		scanf("%d", &r);
	r = all_bcast_i(r);
	on_one
		scanf("%d", &m);
	m = all_bcast_i(m);

    {

		double (*spread A)[l]::[n][r];
		double (*spread B)[l]::[r][m];
		double (*spread C)[l]::[n][m];
		
		/* Allocate and initialize matrices */
		
		A = all_spread_malloc (l*l, n*r*sizeof(double));
		B = all_spread_malloc (l*l, r*m*sizeof(double));
		C = all_spread_malloc (l*l, n*m*sizeof(double));
		{
			int ii, jj, i, j, tmp;
			int x, y;
			
			for_my_2D(i,j,tmp,l,l) {
				
				/* Initialize A matrix */
				for (ii=0; ii<n; ii++)
					for (jj=0; jj<r; jj++) 
						A[i][j][ii][jj] = (i*n+ii)*(j*r+jj);
				
				/* Initialize B matrix */
				for (ii=0; ii<r; ii++)
					for (jj=0; jj<m; jj++) 
						B[i][j][ii][jj] = (i*r+ii)*(j*m+jj);
			}
		}
		
		/* Time the Canon multiply */
		barrier();
		
		secs = get_seconds();
		all_canon_mm(l,n,r,m,A,B,C);
		secs = get_seconds() - secs;
		
		on_one {
			printf("Global Blocked Multiply: %6.4f seconds", secs);
			if (secs > 0.001) 
				printf(" %6.2f MFLOPS (%5.2f per PE)\n",
					   2.0*n*m*r*l*l*l/(1000000.0*secs),
					   2.0*n*m*r*l*l*l/(1000000.0*secs*PROCS));
			else
				printf("\n");
		}
		
		/* Check whether we have the right results. */
		{
			double k2s = 0.0;
			int i,j,ii,jj,k,tmp;
            int failed = 0;
			
			for (k=0;k<r*l;k++) k2s = k2s + k*k;
			
			for_my_2D(i,j,tmp,l,l) {
				for (ii=0;ii<n;ii++)
					for (jj=0;jj<m;jj++) {
						double val = (double)(i*n+ii)*(double)(j*m+jj)*k2s;
						double err = fabs(C[i][j][ii][jj] - val);
						if(err/C[i][j][ii][jj] > 1.0/pow(2,50)) {
                            failed = 1; 
/* chris: commented out printf 
							printf("C[%d][%d]::[%d][%d]=%f, should be %f\n",
								   i, j, ii, jj, C[i][j][ii][jj], val);
*/
						}
					}
			}
			barrier();
            if (failed) printf("Wrong answer\n");
		}
		barrier();
		
		/* Time the local multiply loops without any communication. */
		secs = get_seconds();
		all_fake_mm(l,n,r,m,A,B,C);
		secs = get_seconds() - secs;
		
		on_one {
			printf("Local Blocked Multiply:  %6.4f seconds", secs);
			if (secs > 0.001) 
				printf(" %6.2f MFLOPS (%5.2f per PE)\n",
					   2.0*n*m*r*l*l*l/(1000000.0*secs),
					   2.0*n*m*r*l*l*l/(1000000.0*secs*PROCS));
			else
				printf("\n");
		}
		barrier();
		
		
		/* Time the communication without any local computation */
		secs = get_seconds();
		all_cannon_move(l,n,r,m,A,B,C);
		secs = get_seconds() - secs;
		
		on_one {
			printf("Global Blocked Move:     %6.4f seconds", secs);
			if (secs > 0.001) 
				printf(" %6.2f MBPS   (%5.2f per PE)\n",
					   (n*r+m*r)*l*l*l*sizeof(double)/(1000000.0*secs),
					   (n*r+m*r)*l*l*l*sizeof(double)/(1000000.0*secs*PROCS));
			
			else
				printf("\n");
		}
	}
}
