#include <split-c/split-c.h>

#define min(x,y) (((x)<(y))?(x):(y))
#define max(x,y) (((x)>(y))?(x):(y))
/* 

  Basic Operations on Matrixes

*/

void zero_matrix(int r, int c, double A[r][c])
{
  int i,j;
  for (i=0;i<r;i++)
      for (j=0;j<c;j++) 
	A[i][j] = 0.0;
}

/* Local matrix multiply */
/* There are two versions in this file: BASIC and FULLSPEED.
   The FULLSPEED version has a kernel written in assembly.
*/

#define FULLSPEED

#ifdef BASIC
void  matrix_mult(int n, int r, int m,
	double C[n][m], double A[n][r], double B[r][m])
{
  int i,j,k;
  for (i=0; i<n; i++) {
    for (j=0; j<m; j++) {
      for (k=0; k<r ;k+=1) C[i][j] = C[i][j] + A[i][k]*B[k][j];
    }
  }
}
#endif

/*
  The following implementation of matrix_mult() runs at 6.5 to 7.5
megaflops for matrices ranging in size from 16*16 to 256*256.

  The key to achieving this speed is knowing what the peak
performance.  I call this "PERFORMANCE BY DESIGN."

  The details are as follows: We were losing speed for two reasons: 1)
the inner loop, coded in assembly language was well coded but not well
designed and 2) for matrices larger than 64 by 64, we were not
effectively managing the cache.

  The previous 3x3 matrix multiply included 27 loads and 9 stores for 27
multiply-accumulate operations.  The new 2x2 matrix multiply routine
keeps the values of C[i..i+1][j..j+1] in registers throughout the k
loop.  Thus, we have 8 loads and no stores for 8 multiply-accumulate
operations.  See mm2x2.s for details about how to avoid various hazards
and hence achieve near full performance.

  It is important to avoid routine cache misses in the inner loop.  In
this example, we have C[i][j] += A[i][k] * B[k][j] with i being the 
outer loop, j being the middle loop and k being the inner loop.  If 
B fits entirely inside cache (which it does up to 64 by 128) then we will
have no routine cache misses in the inner loop.  B will remain in the
cache except for a row or two which will get knocked out by A[i][*]
or C[i][*].  Likewise, A[i] and C[i] will remain in cache.  If B is 
slightly too large to fit in cache, B[1..kincr][j] will still fit
in cache and we can preceed as before.  Note that B[1..kincr][j] is
entirely contiguous hence it will fit perfectly in cache.  

  I put in some hooks for playing with jincr when j is so large that
making kincr small enough to fit in cache in not a winning proposition
because then our inner loop is too short.  However, playing with jincr
is a much more tricky thing to get right.  The problem is that
B[1..kincr][1..jincr] may be small enough to fit into cache, but since
it isn't contiguous if m is a power of 2, the cache performance will
not improve.  So, I just set jincr huge.  I left the code in case I or
someone return to this at a later date.  I did some limited testing
and believe that it is correct, though more testing would be necessary 
before it could be released.
*/

#ifdef FULLSPEED
void  matrix_mult(int n, int r, int m,
	double C[n][m], double A[n][r], double B[r][m])
{
    int i,j,k,kmax,kk;
    int n2 = n%2;
    int m2 = m%2;
    int jincr = min(m,819200);
    int kincr = max(8192/(jincr*2),1)*2;
    int jj,jmax;
    int kmin;
    if (kincr <= 8) kincr = 819200;
    
    
    /* Top boundary rows */
    for (i=0; i<n2; i++) 
	for (j=0; j<m; j++) 
	    for (k=0; k<r; k++) 
		C[i][j] = C[i][j] + A[i][k]*B[k][j];
    
    /* Left boundary columns */
    for (j=0; j<m2; j++) 
	for (i=n2; i<n; i++) 
	    for (k=0; k<r; k++) 
		C[i][j] = C[i][j] + A[i][k]*B[k][j];
    
    /* 2x2 blocks on interior */
    for (i=n2; i<n; i+=2) {
	for (j=m2; j<m; j+=2) {
	    for (k=0; k<r%2; k+=1) {
		C[i]  [j]   = C[i  ][j]   + A[i]  [k]*B[k][j];
		C[i+1][j]   = C[i+1][j]   + A[i+1][k]*B[k][j];
		
		C[i]  [j+1] = C[i  ][j+1] + A[i]  [k]*B[k][j+1];
		C[i+1][j+1] = C[i+1][j+1] + A[i+1][k]*B[k][j+1];
	    }
	}
    }
    kmin = k;
    for (jj=m2; jj<m; jj+=jincr) {
	jmax = min(m,jj+jincr);
	for (  kk = kmin; kk<r; kk += kincr) {
	    kmax = min(r,kk+kincr);
	    for (i=n2; i<n; i+=2) {
		for (j=jj; j<jmax; j+=2) {
		    if (kmax > kk) { 
			splitc_poll();
			mm2x2_inner_k(&C[i][j],&A[i][kk],&B[kk][j],8*r,8*m,kmax-kk);
		    }
		}
	    }
	}
    }
}


#endif























