/*									tab:4
 *
 * em3d.sc - Main file for simple version of electromagnetic problem
 *
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: 			Steve Lumetta
 * Version:			25
 * Creation Date:	Sat Jan 23 12:45:18 1993
 * Filename:		em3d.sc
 * History:
 *	ssl	24	Wed Jan 19 15:34:38 1994
 *		changed rpc to atomic
 *	SL	1	Sat Jan 23 12:48:43 1993
 *		Prepared for release as example.
 */

#include <split-c/split-c.h>
#include <split-c/atomic.h>
#include <split-c/com.h>
#include <malloc.h>

#define _EM3D_MAIN
#include "em3d.h"

int n_nodes = 100, d_nodes = 5, local_p = 100, dist_span = 1;

/* all_prepare_packing changes the information used for load-balancing into
   a form more efficient for computation.  */
void all_prepare_packing()
{
    node_t *cur_node;
    int i;
	
    /* For all nodes, replace node pointers with pointers directly to 
       the data values. */
    for (cur_node = h_nodes; cur_node; cur_node = cur_node->next)
        for (i = 0; i < cur_node->from_nodes.length; i++)
			cur_node->from_nodes.nv.values[i] =
	            &cur_node->from_nodes.nv.nodes[i]->value;
    for (cur_node = e_nodes; cur_node; cur_node = cur_node->next)
		for (i = 0; i < cur_node->from_nodes.length; i++)
			cur_node->from_nodes.nv.values[i] =
	            &cur_node->from_nodes.nv.nodes[i]->value;
    barrier();
}

double g_t_time=0;

/* all_compute_e performs a single time step on the e nodes--each node is
   adjusted with each of its from links. */
void all_compute_e()
{
    node_t *cur_node;
    int i;
	double t_time;

	barrier();
	t_time = get_seconds();
	
    for (cur_node = e_nodes; cur_node; cur_node = cur_node->next) {
#if 1
		register double cur_value;

		cur_value = cur_node->value;
		for (i = 0; i < cur_node->from_nodes.length-1;) {
			register double *global p1, *global p2;
			register double d1, d2, d3, d4, d5, d6;
				
			p1 = cur_node->from_nodes.nv.values[i];
			d1 = cur_node->from_nodes.coeffs[i];
			d2 = *p1;
			p2 = cur_node->from_nodes.nv.values[i+1];
			d5 = d1*d2;
			d3 = *p2;
			d4 = cur_node->from_nodes.coeffs[i+1];
			cur_value -= d5;
			d6 = d3*d4;
			i += 2;
			cur_value -= d6;
		}
		if (i == cur_node->from_nodes.length-1) {
			cur_value -= (*cur_node->from_nodes.nv.values[i])*
				cur_node->from_nodes.coeffs[i];
		}

		cur_node->value = cur_value;
		splitc_poll();
#endif
#if 0
		for (i = 0; i < cur_node->from_nodes.length; i++)
			cur_node->value -= (*cur_node->from_nodes.nv.values[i])*
	            cur_node->from_nodes.coeffs[i];
#endif

	}

    barrier();
	t_time = get_seconds() - t_time;

	t_time = all_reduce_to_one_dmax(t_time);
	g_t_time += t_time;
}

/* all_compute_h performs a single time step on the h nodes--each node is
   adjusted with each of its from links. */
void all_compute_h()
{
    node_t *cur_node;
    int i;
	
    for (cur_node = h_nodes; cur_node; cur_node = cur_node->next)
		for (i = 0; i < cur_node->from_nodes.length; i++)
			cur_node->value -= (*cur_node->from_nodes.nv.values[i])*
	            cur_node->from_nodes.coeffs[i];
    barrier();
}

/* dump_nodes frees memory used for nodes which have been taken by other
   processors (free should not be called inside the atomic remove_h_node
   and remove_e_node operations). */
void dump_nodes() 
{
	node_t* free_node;
	
	while (dump_list) {
		free_node = dump_list;
		dump_list = dump_list->next;
		free(free_node->to_nodes.nodes);
		free(free_node->from_nodes.nv.nodes);
		free(free_node->from_nodes.coeffs);
		free(free_node);
	}
}

extern void get_h_addr();
extern void get_e_addr();
/* all_grab_random tests the grab procedures */
void all_grab_random()
{
    node_t *global new_node;
    int from_proc;
    
    from_proc = gen_number(PROCS);
    new_node = toglobal(from_proc, (void *)atomic_i(get_h_addr,
												 toglobal(from_proc, 0)));
    grab_h_node(new_node);
	all_atomic_sync();
	dump_nodes();

    from_proc = gen_number(PROCS);
    new_node = toglobal(from_proc, (void *)atomic_i(get_e_addr,
												 toglobal(from_proc, 0)));
    grab_e_node(new_node);
	all_atomic_sync();
	dump_nodes();
}

splitc_main(int argc, char **argv)
{
    int t;
    int c1;
	double pack_time;

	if (argc > 1)  /* Number of nodes */
		sscanf(argv[1], "%d", &n_nodes);
	if (argc > 2)  /* Degree of nodes */
		sscanf(argv[2], "%d", &d_nodes);
	if (argc > 3)  /* Percent of remote links */
		if (sscanf(argv[3], "%d", &local_p))
			local_p = 100 - local_p;
	if (argc > 4)  /* Distribution span */
		sscanf(argv[4], "%d", &dist_span);

	
    /* First initialize all of the local aliases for spread arrays. */
    cost = tolocal(&g_cost[MYPROC][0]);
    my_cost = &cost[MYPROC];
	
    /* Seed the random number generator. */
    srandom(MYPROC*17);
	
    /* Initialize the graph. */
/*	on_one{printf("Making a graph\n");}*/
    all_initialize_graph();
    c1 = *my_cost;
    
    /* Balance the load. */
/*    if (MYPROC == 0) printf("Scramble...\n");
    all_grab_random();*/
	
    /* Prepare for computation. */
/*    on_one {printf("Preparing to compute...\n");}*/
	barrier();
	pack_time = get_seconds();
    all_prepare_packing ();
    pack_time = get_seconds() - pack_time;
	pack_time = all_reduce_to_one_dmax(pack_time);
/*	on_one{printf("Prepare: %.6lf seconds.\n", pack_time);}
	barrier();*/
    
    /* Compute! */
    for (t = 0; t < TIME_STEPS; t++) {
/*        on_one{printf("Time step %d\n",t + 1);}
		barrier();*/
        all_compute_e();
		all_compute_h();
    }
	
	on_one {
		int i = n_nodes*d_nodes;
		
/*		printf("        Total\n");*/
		printf("%d %14.6lf%15.6f%15.6lf%15.6lf\n",
			   100 - local_p,
			   (1000000*pack_time)/i,
			   0.0,
			   (1000000*g_t_time)/(TIME_STEPS*i),
			   (1000000*g_t_time)/(TIME_STEPS*i));
	}

/*    on_one{printf("Finished.\n");}*/
}




