/*									tab:4
 *
 * em3d.sc - Main file for the electromagnetic problem in three dimensions.
 *
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: 			Steve Lumetta
 * Version:			84
 * Creation Date:	Mon Jan 25 11:58:17 1993
 * Filename:		em3d.sc
 * History:
 *	ssl	83	Wed Jan 19 15:34:30 1994
 *		changed rpc to atomic
 *	SL	1	Mon Jan 25 11:59:00 1993
 *		Prepared for release as example.
 */

#include <split-c/split-c.h>
#include <split-c/atomic.h>
#include <split-c/com.h>
#include <malloc.h>

#define _EM3D_MAIN
#include "em3d.h"

int n_nodes = 100, d_nodes = 5, local_p = 100, dist_span = 1;

/* all_prepare_packing changes the information used for load-balancing into
   a form more efficient for computation.  Since many of the loops contain
   all-to-all communication, each processor starts with it's 'neighbor,'
   avoiding communication jams caused when all processors try to communicate
   with a single processor. */
void all_prepare_packing ()
{
    int i, n, j, dest_proc;
	node_t *cur_node;
    DATA *package;

    /* For all other processors, count the number of h and e nodes which 
       must be sent--store this information locally and on the processor
       to which you will send the nodes. */
	for (i = 0; i < PROCS; i++) {
		h_pack_area[i] = NULL;
		h_pack_len[i] = 0;
		e_pack_area[i] = NULL;
		e_pack_len[i] = 0;
	}
	for (cur_node = h_nodes; cur_node; cur_node = cur_node->next) {
		for (i = 0; i < cur_node->to_nodes.length; i++) {
			dest_proc = toproc(cur_node->to_nodes.nodes[i]);
			if (h_pack_area[dest_proc] == (void *)cur_node)
				continue;
			else {
				h_pack_len[dest_proc]++;
				h_pack_area[dest_proc] = (void *)cur_node;
			}
		}
	}
    for (n = 0, i = (MYPROC + 1)%PROCS ; n < PROCS - 1;
		 n++, i = (i + 1)%PROCS)
		g_h_rcv_len[i][MYPROC] :- h_pack_len[i];
	for (cur_node = e_nodes; cur_node; cur_node = cur_node->next) {
		for (i = 0; i < cur_node->to_nodes.length; i++) {
			dest_proc = toproc(cur_node->to_nodes.nodes[i]);
			if (e_pack_area[dest_proc] == (void *)cur_node)
				continue;
			else {
				e_pack_len[dest_proc]++;
				e_pack_area[dest_proc] = (void *)cur_node;
			}
		}
	}
    for (n = 0, i = (MYPROC + 1)%PROCS ; n < PROCS - 1;
		 n++, i = (i + 1)%PROCS)
		g_e_rcv_len[i][MYPROC] :- e_pack_len[i];
    store_sync (2*(PROCS - 1)*sizeof (int));
    /* Note that since the counters for store are unique, we must make
       sure that none of the data received in later stages allows a
       processor to falsely pass store_sync.  Hence the barrier. */
    barrier ();	
	
    /* Create packing lists and storage areas for sending nodes to other
       processors.  Also create areas to receive nodes from other
       processors and propagate global pointers to these areas to the
       senders.  Count the total number of nodes received. */
    h_rcv_total = e_rcv_total = 0;
    for (n = 0, i = (MYPROC + 1)%PROCS ; n < PROCS - 1; 
		 n++, i = (i + 1)%PROCS) {
        h_rcv_total += h_rcv_len[i];
        h_pack_list[i] = (DATA **)malloc (h_pack_len[i]*sizeof (DATA *));
        h_pack_area[i] = (DATA *)malloc (h_pack_len[i]*sizeof (DATA));
		h_rcv_area[i] = (DATA *)malloc (h_rcv_len[i]*sizeof (DATA));
		h_pack_len[i] = 0;
		g_h_put_area[i][MYPROC] :- (DATA *global)h_rcv_area[i];
        e_rcv_total += e_rcv_len[i];
        e_pack_list[i] = (DATA **)malloc (e_pack_len[i]*sizeof (DATA *));
        e_pack_area[i] = (DATA *)malloc (e_pack_len[i]*sizeof (DATA));
		e_rcv_area[i] = (DATA *)malloc (e_rcv_len[i]*sizeof (DATA));
		g_e_put_area[i][MYPROC] :- (DATA *global)e_rcv_area[i];
		e_pack_len[i] = 0;
    }
    store_sync (2*(PROCS - 1)*sizeof (DATA *global));
    barrier ();

    /* Fill in the packing lists for other processors and send a copy in
       the pack_area so that the receiver can properly adjust from links
       to point into the data reception area. */
	for (cur_node = h_nodes; cur_node; cur_node = cur_node->next) {
		for (i = 0; i < cur_node->to_nodes.length; i++) {
			dest_proc = toproc(cur_node->to_nodes.nodes[i]);
			if (dest_proc == MYPROC)
				continue;
			if (h_pack_list[dest_proc][h_pack_len[dest_proc] - 1] ==
				(void *)&cur_node->value)
				continue;
			else {
				h_pack_list[dest_proc][h_pack_len[dest_proc]] =
					&cur_node->value;
				*(node_t **)&h_pack_area[dest_proc][h_pack_len[dest_proc]] =
					cur_node;
				h_pack_len[dest_proc]++;
			}
		}
	}
	for (n = 0, i = (MYPROC + 1)%PROCS ; n < PROCS - 1; 
		 n++, i = (i + 1)%PROCS)
		if (h_pack_len[i])
			bulk_store(h_put_area[i], h_pack_area[i],
					   h_pack_len[i]*sizeof(DATA));
	for (cur_node = e_nodes; cur_node; cur_node = cur_node->next) {
		for (i = 0; i < cur_node->to_nodes.length; i++) {
			dest_proc = toproc(cur_node->to_nodes.nodes[i]);
			if (dest_proc == MYPROC)
				continue;
			if (e_pack_list[dest_proc][e_pack_len[dest_proc] - 1] ==
				(void *)&cur_node->value)
				continue;
			else {
				e_pack_list[dest_proc][e_pack_len[dest_proc]] =
					&cur_node->value;
				*(node_t **)&e_pack_area[dest_proc][e_pack_len[dest_proc]] =
					cur_node;
				e_pack_len[dest_proc]++;
			}
		}
	}
	for (n = 0, i = (MYPROC + 1)%PROCS ; n < PROCS - 1; 
		 n++, i = (i + 1)%PROCS)
		if (e_pack_len[i])
			bulk_store(e_put_area[i], e_pack_area[i],
					   e_pack_len[i]*sizeof(DATA));
	store_sync (sizeof(DATA)*(h_rcv_total + e_rcv_total));
	barrier();
	
	h_ghost_nodes = NULL;
	if (h_rcv_total) {
		h_ghost_nodes =
			(ghost_node_t *)malloc(h_rcv_total*sizeof(ghost_node_t));
		for (i = 0; i < h_rcv_total - 1; i++)
			h_ghost_nodes[i].next = h_ghost_nodes + i + 1;
		h_ghost_nodes[i].next = NULL;
		for (n = 0, i = (MYPROC + 1)%PROCS, j = 0 ; n < PROCS - 1; 
			 n++, i = (i + 1)%PROCS) {
			int k;
			
			for (k = 0, package = h_rcv_area[i]; k < h_rcv_len[i];
				 k++, package++, j++)
				h_ghost_nodes[j].rval =
					(DATA *global)toglobal(i, *(node_t **)package);
		}
	}
	
	e_ghost_nodes = NULL;
	if (e_rcv_total) {
		e_ghost_nodes =
			(ghost_node_t *)malloc(e_rcv_total*sizeof(ghost_node_t));
		for (i = 0; i < e_rcv_total - 1; i++)
			e_ghost_nodes[i].next = e_ghost_nodes + i + 1;
		e_ghost_nodes[i].next = NULL;
		for (n = 0, i = (MYPROC + 1)%PROCS, j = 0 ; n < PROCS - 1; 
			 n++, i = (i + 1)%PROCS) {
			int k;
			
			for (k = 0, package = e_rcv_area[i]; k < e_rcv_len[i];
				 k++, package++, j++)
				e_ghost_nodes[j].rval =
					(DATA *global)toglobal(i, *(node_t **)package);
		}	
	}
	
    /* For all nodes, adjust the from links to point to the data value of
       the node (if local) or the location in the data reception area where
       the node's owner will place the value (if remote).
	   
       Notice that the from_nodes nodes pointers are overwritten gradually as
       the local values pointers are filled in.  Since the local pointers are
       smaller than the global nodes pointers, no problems arise if they are
       overwritten in increasing order. */
	for (cur_node = h_nodes; cur_node; cur_node = cur_node->next) {
		for (j = 0; j < cur_node->from_nodes.length; j++) {
			dest_proc = toproc(cur_node->from_nodes.nv.nodes[j]);
			if (dest_proc == MYPROC)
				cur_node->from_nodes.nv.values[j] =
					&((node_t *)cur_node->from_nodes.nv.nodes[j])->value;
			else {
				for (i = 0; ; i++)
					if (cur_node->from_nodes.nv.nodes[j] == 
						(node_t *global)e_ghost_nodes[i].rval)
						break;
				cur_node->from_nodes.nv.values[j] =
					&e_ghost_nodes[i].value;
			}
		}
    }
	for (cur_node = e_nodes; cur_node; cur_node = cur_node->next) {
		for (j = 0; j < cur_node->from_nodes.length; j++) {
			dest_proc = toproc(cur_node->from_nodes.nv.nodes[j]);
			if (dest_proc == MYPROC)
				cur_node->from_nodes.nv.values[j] =
					&((node_t *)cur_node->from_nodes.nv.nodes[j])->value;
			else {
				for (i = 0; ; i++)
					if (cur_node->from_nodes.nv.nodes[j] == 
						(node_t *global)h_ghost_nodes[i].rval)
						break;
				cur_node->from_nodes.nv.values[j] =
					&h_ghost_nodes[i].value;
			}
		}
    }

	for (i = 0; i < h_rcv_total; i++)
		h_ghost_nodes[i].rval =
			&((node_t *global)h_ghost_nodes[i].rval)->value;
	for (i = 0; i < e_rcv_total; i++)
		e_ghost_nodes[i].rval =
			&((node_t *global)e_ghost_nodes[i].rval)->value;

    barrier ();
}

double g_p_time = 0, g_c_time = 0, g_t_time = 0;

/* all_compute_e performs a single time step on the e nodes.  Values for
   h nodes are packed up and sent to the processors which require them, 
   then when all of the data has arrived, each node is adjusted
   with each of its from links. */
void all_compute_e ()
{
    int i, j, n;
	node_t* cur_node;
	ghost_node_t *g_node;
	double s_time, p_time, c_time, t_time;

	barrier();
	s_time = get_seconds();

	for (g_node = h_ghost_nodes; g_node; g_node = g_node->next)
			g_node->value = *(g_node->rval);

	p_time = get_seconds() - s_time;
	
	for (cur_node = e_nodes; cur_node; cur_node = cur_node->next) {

#ifdef OPTIMIZE
		double cur_value;

		cur_value = cur_node->value;
		for (i = 0; i < cur_node->from_nodes.length-1;) {
			register double *p1, *p2;
			register double d1, d2, d3, d4, d5, d6;
				
			p1 = cur_node->from_nodes.nv.values[i];
			d1 = cur_node->from_nodes.coeffs[i];
			d2 = *p1;
			p2 = cur_node->from_nodes.nv.values[i+1];
			d5 = d1*d2;
			d3 = *p2;
			d4 = cur_node->from_nodes.coeffs[i+1];
			cur_value -= d5;
			d6 = d3*d4;
			i += 2;
			cur_value -= d6;
		}
		if (i == cur_node->from_nodes.length-1) {
			cur_value -= (*cur_node->from_nodes.nv.values[i])*
				cur_node->from_nodes.coeffs[i];
		}
		cur_node->value = cur_value;
#endif

#ifndef OPTIMIZE
		for (j = 0; j < cur_node->from_nodes.length; j += 1)
			cur_node->value -= (*cur_node->from_nodes.nv.values[j])*
	            cur_node->from_nodes.coeffs[j];

#endif

		splitc_poll();

	}
	
	c_time = get_seconds() - p_time - s_time;
    barrier ();
	t_time = get_seconds() - s_time;

	p_time = all_reduce_to_one_dmax(p_time);
	c_time = all_reduce_to_one_dmax(c_time);
	t_time = all_reduce_to_one_dmax(t_time);
	g_p_time += p_time;
	g_c_time += c_time;
	g_t_time += t_time;
}

/* all_compute_h performs a single time step on the h nodes.  Values for
   e nodes are packed up and sent to the processors which require them, 
   then when all of the data has arrived, each node is adjusted
   with each of its from links. */
void all_compute_h ()
{
    int i, j, n;
	node_t* cur_node;
	ghost_node_t *g_node;

	for (g_node = e_ghost_nodes; g_node; g_node = g_node->next)
		g_node->value = *(g_node->rval);
	
	for (cur_node = h_nodes; cur_node; cur_node = cur_node->next) {
		splitc_poll();
		for (j = 0; j < cur_node->from_nodes.length; j++)
			cur_node->value -= (*cur_node->from_nodes.nv.values[j])*
	            cur_node->from_nodes.coeffs[j];
	}
	
    barrier ();
}

/* dump_nodes frees memory used for nodes which have been taken by other
   processors (free should not be called inside the atomic remove_h_node
   and remove_e_node operations). */
void dump_nodes() 
{
	node_t* free_node;
	
	while (dump_list) {
		free_node = dump_list;
		dump_list = dump_list->next;
		free(free_node->to_nodes.nodes);
		free(free_node->from_nodes.nv.nodes);
		free(free_node->from_nodes.coeffs);
		free(free_node);
	}
}

extern void get_h_addr();
extern void get_e_addr();
/* all_grab_random tests the grab procedures */
void all_grab_random()
{
    node_t *global new_node;
    int from_proc;
    
    from_proc = gen_number(PROCS);
    new_node = toglobal(from_proc, (void *)atomic_i(get_h_addr,
												 toglobal(from_proc, 0)));
    grab_h_node(new_node);
	all_atomic_sync();
	dump_nodes();

    from_proc = gen_number(PROCS);
    new_node = toglobal(from_proc, (void *)atomic_i(get_e_addr,
												 toglobal(from_proc, 0)));
    grab_e_node(new_node);
	all_atomic_sync();
	dump_nodes();
}

splitc_main (int argc, char **argv)
{
    int t;
    int c1;
	double pack_time;

	if (argc > 1)  /* Number of nodes */
		sscanf(argv[1], "%d", &n_nodes);
	if (argc > 2)  /* Degree of nodes */
		sscanf(argv[2], "%d", &d_nodes);
	if (argc > 3)  /* Percent of remote links */
		if (sscanf(argv[3], "%d", &local_p))
			local_p = 100 - local_p;
	if (argc > 4)  /* Distribution span */
		sscanf(argv[4], "%d", &dist_span);

	
    /* First initialize all of the local aliases for spread arrays. */
    cost = tolocal (&g_cost[MYPROC][0]);
    my_cost = &cost[MYPROC];
    h_rcv_len = tolocal (&g_h_rcv_len[MYPROC][0]);
    h_put_area = tolocal (&g_h_put_area[MYPROC][0]);
    e_rcv_len = tolocal (&g_e_rcv_len[MYPROC][0]);
    e_put_area = tolocal (&g_e_put_area[MYPROC][0]);
	
    /* Seed the random number generator. */
    srandom (MYPROC*17);
	
    /* Initialize the graph. */
/*	on_one{printf("Making a graph\n");}*/
    all_initialize_graph ();
    c1 = *my_cost;
	
    /* Balance the load. */
/*    if (MYPROC == 0) printf("Scramble...\n");
    all_grab_random ();*/
	
    /* Prepare for computation. */
/*    on_one {printf("Preparing to compute...\n");}*/
	barrier();
	pack_time = get_seconds();
    all_prepare_packing ();
    pack_time = get_seconds() - pack_time;
	pack_time = all_reduce_to_one_dmax(pack_time);
/*	on_one{printf("Prepare: %.6lf seconds.\n", pack_time);}
	barrier();*/
	
    /* Compute! */
	for (t = 0; t < TIME_STEPS; t++) {
/*        on_one{printf("Time step %d\n",t + 1);}
		barrier();*/
        all_compute_e();
		all_compute_h();
    }

	on_one {
		int i = n_nodes*d_nodes;
		
/*		printf("        Pack          Compute         Total\n");*/
		printf("%d %14.6lf%15.6lf%15.6lf%15.6lf\n",
			   100 - local_p,
			   (1000000*pack_time)/i,
			   (1000000*g_p_time)/(TIME_STEPS*i),
			   (1000000*g_c_time)/(TIME_STEPS*i),
			   (1000000*g_t_time)/(TIME_STEPS*i));
	}

/*    on_one{printf("Finished.\n");}*/
}


