/*									tab:4
 *
 * make_graph.sc - Create a graph to be balanced and solved for
 *                 the version of the electromagnetic problem with
 *                 store used for transferring data.  
 *
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Steve Lumetta
 * Version:			5
 * Creation Date:	Sat Jan 23 12:51:07 1993
 * Filename:		make_graph.sc
 * History:
 *	ssl	5	Wed Jan 19 15:33:18 1994
 *		changed rpc to atomic. rpc not supported in libsplit-c version 125
 *	SL	1	Sat Jan 23 12:51:45 1993
 *		Prepared for release as example.
 */

#include <split-c/split-c.h>
#include <split-c/atomic.h>
#include <malloc.h>

#include "em3d.h"

extern int n_nodes, d_nodes, local_p, dist_span;

/* These constants determine the nature of the random graph. */
#define NUM_H_NODES n_nodes /* 100 */
#define H_NODE_DEGREE d_nodes /* 5 */
#define H_PERCENT_LOCAL local_p /* 0 */

#define NUM_E_NODES n_nodes /* 100 */
#define E_NODE_DEGREE d_nodes /* 5 */
#define E_PERCENT_LOCAL local_p /* 100 */

#define DIST_SPAN dist_span /* 6 */

node_t **h_table;
node_t **e_table;

/* inc_from is an atomic procedure which increments the length of the
   from_node list of a node. */
static void inc_from(node_t *a_node)
{
	(a_node->from_nodes.length)++;
}

/* add_from is an atomic procedure which adds a node to the from_node list
   of a node. */
static void add_from(node_t *dest_node, node_t *global orig_node)
{
	dest_node->from_nodes.nv.nodes[dest_node->from_nodes.length++] = 
		orig_node;
}

/* This is hacked to speed all_initialize graph--DON'T use it
   for moving nodes around randomly! */
/* get_h_addr is an rpc call returning the nth h_node on a processor. */
void get_h_addr(int proc, int num)
{
/*	node_t *ret_node = h_nodes;
	
	while (num-- > 0) ret_node = ret_node->next;
	
	atomic_return_i (ret_node);*/
	atomic_return_i (h_table[num]);
}

/* This is hacked to speed all_initialize graph--DON'T use it
   for moving nodes around randomly! */
/* get_e_addr is an rpc call returning the nth e_node on a processor. */
void get_e_addr(int proc, int num)
{
/*	node_t *ret_node = e_nodes;
	
	while (num-- > 0) ret_node = ret_node->next;
	
	atomic_return_i (ret_node);*/
	atomic_return_i (e_table[num]);
}

/* all_initialize_graph creates a random graph to be solved. */
void all_initialize_graph()
{
	int i, j, k, dest_proc;
	node_t *cur_node;
	node_t *global other_node;

	/* First the nodes are allocated and initializes. */
	n_h_nodes = NUM_H_NODES;
	h_table = (node_t **)malloc(NUM_H_NODES*sizeof(node_t *));
	h_table[0] = cur_node = h_nodes = (node_t *)malloc(sizeof(node_t));
	cur_node->prev = &h_nodes;
	for (i = 0; i < n_h_nodes - 1; i++) {
		cur_node->locked = 0;
		cur_node->from_nodes.length = 0;
		h_table[i  + 1] = cur_node->next = 
			(node_t *)malloc(sizeof(node_t));
		cur_node->next->prev = &cur_node->next;
		cur_node = cur_node->next;
	}
	cur_node->locked = 0;
	cur_node->from_nodes.length = 0;
	cur_node->next = NULL;
	
	n_e_nodes = NUM_E_NODES;
	e_table = (node_t **)malloc(NUM_E_NODES*sizeof(node_t *));
	e_table[0] = cur_node = e_nodes = (node_t *)malloc(sizeof(node_t));
	cur_node->prev = &e_nodes;
	for (i = 0; i < n_e_nodes - 1; i++) {
		cur_node->locked = 0;
		cur_node->from_nodes.length = 0;
		e_table[i + 1] = cur_node->next = (node_t *)malloc(sizeof(node_t));
		cur_node->next->prev = &cur_node->next;
		cur_node = cur_node->next;
	}
	cur_node->locked = 0;
	cur_node->from_nodes.length = 0;
	cur_node->next = NULL;
	
	barrier();
	
	/* Next the lists of dependent nodes are filled in and the
	   dependents informed so that the length of the from_nodes
	   lists is known later. */
	for (cur_node = h_nodes; cur_node; cur_node = cur_node->next) {
		cur_node->to_nodes.length = H_NODE_DEGREE;
		cur_node->to_nodes.nodes = (node_t *global *)
			malloc(H_NODE_DEGREE*sizeof (node_t *global));
		for (j = 0; j < H_NODE_DEGREE; j++) {
			do {
				dest_proc = (check_percent(H_PERCENT_LOCAL) ? 
							 MYPROC :
							 (MYPROC + PROCS + gen_signed_number(DIST_SPAN)) % PROCS);

				other_node = toglobal(dest_proc, (node_t*)
									  atomic_i(get_e_addr, toglobal(dest_proc,
											gen_number(n_e_nodes))));
				for (k = 0; k < j; k++)
					if (other_node == cur_node->to_nodes.nodes[k])
						break;
			} while (k < j);
			cur_node->to_nodes.nodes[j] = other_node;
			atomic(inc_from,other_node);
		}
	}    
	for (cur_node = e_nodes; cur_node; cur_node = cur_node->next) {
		cur_node->to_nodes.length = E_NODE_DEGREE;
		cur_node->to_nodes.nodes = (node_t *global *)
			malloc(E_NODE_DEGREE*sizeof(node_t *global));
		for (j = 0; j < E_NODE_DEGREE; j++) {
			do {
				dest_proc = (check_percent(E_PERCENT_LOCAL) ? 
							 MYPROC :
							 (MYPROC + PROCS + gen_signed_number(DIST_SPAN)) % PROCS);
							 
				other_node = toglobal(dest_proc, (node_t*)
									  atomic_i(get_h_addr, toglobal(dest_proc,
											gen_number(n_h_nodes))));
				for (k = 0; k < j; k++)
					if (other_node == cur_node->to_nodes.nodes[k])
						break;
			} while (k < j);
			cur_node->to_nodes.nodes[j] = other_node;
			atomic(inc_from,other_node);
		}
	}    
	all_atomic_sync();
	
	/* Using the length determined in the last section, space is
	   allocated for links to nodes on which a node depends and for
	   coefficients for the dependencies. */
	for (cur_node = h_nodes; cur_node; cur_node = cur_node->next) {
		cur_node->from_nodes.nv.nodes = (node_t *global *)
			malloc(cur_node->from_nodes.length*sizeof(node_t *global));
		cur_node->from_nodes.coeffs = (DATA *)
			malloc(cur_node->from_nodes.length*sizeof(DATA));
		cur_node->from_nodes.length = 0;
	}
	for (cur_node = e_nodes; cur_node; cur_node = cur_node->next) {
		cur_node->from_nodes.nv.nodes = (node_t *global *)
			malloc(cur_node->from_nodes.length*sizeof(node_t *global));
		cur_node->from_nodes.coeffs = (DATA *)
			malloc(cur_node->from_nodes.length*sizeof(DATA));
		cur_node->from_nodes.length = 0;
	}
	barrier();
	
	/* The nodes are then added into the from_nodes lists. */
	for (cur_node = h_nodes; cur_node; cur_node = cur_node->next)
		for (j = 0; j < cur_node->to_nodes.length; j++)
			atomic(add_from, cur_node->to_nodes.nodes[j],
				   (node_t *global)cur_node);
	for (cur_node = e_nodes; cur_node; cur_node = cur_node->next)
		for (j = 0; j < cur_node->to_nodes.length; j++)
			atomic(add_from, cur_node->to_nodes.nodes[j],
				   (node_t *global)cur_node);
	all_atomic_sync();
	
	/* Finally, the cost of computation is calculated for each node
	   to aid in balancing the load. */
	*my_cost = 0;	
	for (cur_node = h_nodes; cur_node; cur_node = cur_node->next)
		for (j = 0; j < cur_node->from_nodes.length; j++)
			*my_cost += ((toproc(cur_node->from_nodes.nv.nodes[j]) ==
						  MYPROC) ? COST_LOCAL : COST_REMOTE);
	for (cur_node = e_nodes; cur_node; cur_node = cur_node->next)
		for (j = 0; j < cur_node->from_nodes.length; j++)
			*my_cost += ((toproc(cur_node->from_nodes.nv.nodes[j]) ==
						  MYPROC) ? COST_LOCAL : COST_REMOTE);
	barrier();
}

