/*									tab:4
 *
 * move_node.sc - Moves nodes between processors in the version of
 *                the electromagnetic problem with store used for 
 *                transferring data		
 *
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: 			Steve Lumetta
 * Version:			5
 * Creation Date:	Sat Jan 23 12:49:19 1993
 * Filename:		move_node.sc
 * History:
 *	SL	1	Sat Jan 23 12:49:54 1993
 *		Prepared for release as example.
 */
#include <split-c/split-c.h>
#include <split-c/atomic.h>
#include <malloc.h>

#include "em3d.h"

/* check_percent returns 1 chance percent of the time--otherwise it 
   returns 0. */
int check_percent(int chance)
{
	return (((random()/128)%100) < chance);
}

/* gen_number generates a random number between 0 and limit - 1. */
int gen_number(int limit)
{
	return (random()/128)%limit;
}

/* generates a number between -(limit - 1) and (limit - 1) excluding 0 */
int gen_signed_number(int limit)
{
	int num = gen_number (2*limit);

	return (num < limit ? (num - limit) : (num - limit + 1));
}

/* get_cost returns a hint at another processor's cost function.  If
   there is no hint, the actual value is obtained with a global 
   reference and stored for later use. */
int get_cost(int proc_num)
{
	if (!cost[proc_num])
		return (cost[proc_num] = g_cost[proc_num][proc_num]);
	return cost[proc_num];
}

/* change_from_link is an atomic procedure used to change a single from
   link on a node to point to another global location.  Cost is also
   adjusted. */
static void change_from_link(node_t *a_node, void *global old_addr,
							 void *global new_addr)
{
	int i;
	
	for (i = 0; ; i++)
		if (a_node->from_nodes.nv.nodes[i] == old_addr) {
			a_node->from_nodes.nv.nodes[i] = new_addr;
			break;
		}
	
	*my_cost += (toproc(new_addr) == MYPROC ? COST_LOCAL : COST_REMOTE) -
		(toproc(old_addr) == MYPROC ? COST_LOCAL : COST_REMOTE);
}

/* change_to_link is an atomic procedure used to change a single to
   link on a node to point to another global location. */
static void change_to_link(node_t *a_node, void *global old_addr,
						   void *global new_addr)
{
	int i;
	
	for (i = 0; ; i++)
		if (a_node->to_nodes.nodes[i] == old_addr) {
			a_node->to_nodes.nodes[i] = new_addr;
			break;
		}
}

/* remove_h_node is an atomic procedure used to remove an h node from
   a processor after another processor has obtained all of the information
   it needs to grab the node.   Cost information is also adjusted. */
static void remove_h_node(node_t *a_node)
{
	int i;
	
	for (i = 0; i < a_node->from_nodes.length; i++)
		*my_cost -= (toproc(a_node->from_nodes.nv.nodes[i]) == MYPROC ? 
					 COST_LOCAL : COST_REMOTE);
	*a_node->prev = a_node->next;
	a_node->next = dump_list;
	dump_list = a_node;
	n_h_nodes--;
}

/* remove_e_node is an atomic procedure used to remove an e node from
   a processor after another processor has obtained all of the information
   it needs to grab the node.   Cost information is also adjusted. */
static void remove_e_node(node_t *a_node)
{
	int i;
	
	for (i = 0; i < a_node->from_nodes.length; i++)
		*my_cost -= (toproc(a_node->from_nodes.nv.nodes[i]) == MYPROC ? 
					 COST_LOCAL : COST_REMOTE);
	*a_node->prev = a_node->next;
	a_node->next = dump_list;
	dump_list = a_node;
	n_e_nodes--;
}

/* grab_h_node is the procedure used to obtain an h node known by a global
   pointer.  A claim on the node is asserted, and, if successful, the
   node is moved from the previous owner to the processor which claimed
   it. */
void grab_h_node(node_t *global a_node)
{
	int other_proc = toproc(a_node);
	node_t *new_node;
	int i;
	node_t *global *global copy_from;
	node_t *global *global copy_to;
	DATA *global copy_coeffs;
	
	/* Assert claim on node */
	if (atomic_i(test_and_set_atomic, &a_node->locked))
		return;
	
	/* Make a space for the node */
	new_node = (node_t*)malloc(sizeof(node_t));
	n_h_nodes++;
	
	printf ("H sent from %02X:%08X to %02X:%08X\n", other_proc,
			tolocal(a_node), MYPROC, new_node);
	
	/* Copy the node and tell the old owner to proceed with removal. */
	new_node->locked = 0;
	new_node->value = a_node->value;
	new_node->from_nodes.length = a_node->from_nodes.length;
	new_node->to_nodes.length = a_node->to_nodes.length;
	copy_from = toglobal(other_proc, a_node->from_nodes.nv.nodes);
	copy_to = toglobal(other_proc, a_node->to_nodes.nodes);
	copy_coeffs = toglobal(other_proc, a_node->from_nodes.coeffs);
	new_node->from_nodes.nv.nodes = (node_t *global *)
		malloc(new_node->from_nodes.length*sizeof(node_t *global));
	bulk_read(new_node->from_nodes.nv.nodes, copy_from,
			   new_node->from_nodes.length*sizeof(node_t *global));
	new_node->to_nodes.nodes = (node_t *global *)
		malloc(new_node->to_nodes.length*sizeof(node_t *global));
	bulk_read(new_node->to_nodes.nodes, copy_to,
			   new_node->to_nodes.length*sizeof(node_t *global));
	new_node->from_nodes.coeffs = (DATA *)
		malloc(new_node->from_nodes.length*sizeof(DATA));
	bulk_read(new_node->from_nodes.coeffs, copy_coeffs,
			   new_node->from_nodes.length*sizeof(DATA));
	atomic(remove_h_node,a_node); 
	
	/* For each to link, inform the linked node that this node has moved. */
	for (i = 0; i < new_node->to_nodes.length; i++)
		atomic_long(change_from_link, new_node->to_nodes.nodes[i],
					 a_node, (node_t* global)new_node);
	
	/* For each from link, adjust cost and inform the linked node that
	   this node has moved.   Unlock the node in case someone else now
	   wants to grab it. */
	for (i = 0; i < new_node->from_nodes.length; i++) {
		atomic_long(change_to_link, new_node->from_nodes.nv.nodes[i],
					 a_node, (node_t* global)new_node);
		*my_cost += (toproc(new_node->from_nodes.nv.nodes[i]) == MYPROC ? 
					 COST_LOCAL : COST_REMOTE);
	}
	
	/* Add new node to h_nodes list. */
	if (h_nodes) h_nodes->prev = &new_node->next;
	new_node->prev = &h_nodes;
	new_node->next = h_nodes;
	h_nodes = new_node;
	
	/* Exchange costs with the previous owner to update knowledge. */
	cost[other_proc] = g_cost[other_proc][other_proc];
	g_cost[other_proc][MYPROC] = *my_cost;
}

/* grab_e_node is the procedure used to obtain an e node known by a global
   pointer.  A claim on the node is asserted, and, if successful, the
   node is moved from the previous owner to the procnessor which claimed
   it. */
void grab_e_node(node_t *global a_node)
{
	int other_proc = toproc(a_node);
	node_t *new_node;
	int i;
	node_t *global *global copy_from;
	node_t *global *global copy_to;
	DATA *global copy_coeffs;
	
	/* Assert claim on node */
	if (atomic_i(test_and_set_atomic, &a_node->locked))
		return;
	
	/* Make a space for the node */
	new_node = (node_t*)malloc(sizeof(node_t));
	n_e_nodes++;
	
	printf ("E sent from %02X:%08X to %02X:%08X\n", other_proc,
			tolocal(a_node), MYPROC, new_node);
	
	/* Copy the node and tell the old owner to proceed with removal. */
	new_node->locked = 0;
	new_node->value = a_node->value;
	new_node->from_nodes.length = a_node->from_nodes.length;
	new_node->to_nodes.length = a_node->to_nodes.length;
	copy_from = toglobal(other_proc, a_node->from_nodes.nv.nodes);
	copy_to = toglobal(other_proc, a_node->to_nodes.nodes);
	copy_coeffs = toglobal(other_proc, a_node->from_nodes.coeffs);
	new_node->from_nodes.nv.nodes = (node_t *global *)
		malloc(new_node->from_nodes.length*sizeof(node_t *global));
	bulk_read(new_node->from_nodes.nv.nodes, copy_from,
			   new_node->from_nodes.length*sizeof(node_t *global));
	new_node->to_nodes.nodes = (node_t *global *)
		malloc(new_node->to_nodes.length*sizeof(node_t *global));
	bulk_read(new_node->to_nodes.nodes, copy_to,
			   new_node->to_nodes.length*sizeof(node_t *global));
	new_node->from_nodes.coeffs = (DATA *)
		malloc(new_node->from_nodes.length*sizeof(DATA));
	bulk_read(new_node->from_nodes.coeffs, copy_coeffs,
			   new_node->from_nodes.length*sizeof(DATA));
	atomic(remove_e_node, a_node); 
	
	/* For each to link, inform the linked node that this node has moved. */
	for (i = 0; i < new_node->to_nodes.length; i++)
		atomic_long(change_from_link, new_node->to_nodes.nodes[i],
					 a_node, (node_t *global)new_node);
	
	/* For each from link, adjust cost and inform the linked node that
	   this node has moved.   Unlock the node in case someone else now
	   wants to grab it. */
	for (i = 0; i < new_node->from_nodes.length; i++) {
		atomic_long(change_to_link, new_node->from_nodes.nv.nodes[i],
					 a_node, (node_t *global)new_node);
		*my_cost += (toproc(new_node->from_nodes.nv.nodes[i]) == MYPROC ? 
					 COST_LOCAL : COST_REMOTE);
	}
	
	/* Add new node to e_nodes list. */
	if (e_nodes) e_nodes->prev = &new_node->next;
	new_node->prev = &e_nodes;
	new_node->next = e_nodes;
	e_nodes = new_node;
	
	/* Exchange costs with the previous owner to update knowledge. */
	cost[other_proc] = g_cost[other_proc][other_proc];
	g_cost[other_proc][MYPROC] = *my_cost;
}

