/*									tab:4
 *
 * radix.sc - Split-C sample radix sort program
 *
 * "Copyright (c) 1993 by The Regents of the University of California.
 * All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: 			Rich Martin
 * Version:			48
 * Creation Date:	Fri Jan 22 12:11:57 1993
 * Filename:		radix.sc
 * History:
 *	ssl	48	Thu Mar  3 11:53:34 1994
 *		fixed dependency in all_scan_buckets and changed CMAM to splitc
 *	SL	2	Mon Jan 25 12:13:50 1993
 *		Expanded error reporting for easy automated testing.
 *	SL	1	Fri Jan 22 12:13:02 1993
 *		Prepared for release as example.
 */

#include <split-c/split-c.h>
#include <split-c/control.h>
#include <split-c/com.h>

#define MAX_KEYS_PER_PROC  1048576 /* default settings */
#define LOG_KEYS 20

#define BITS_PER_DIGIT 16    /* bits per digit used in sort */
#define RADIX 65536           /* radix (digit size)--must be power of 2 */
#define DIGITS_PER_KEY 2      /* digits in a key (round up) */

#define MIN(A,B) (A < B ? A : B)

int keys_per_proc; /* keys per processor--must be power of 2 */
int log_keys;       

/* Two key lists are used--on each pass the keys are moved from one to the
   other, sorting a specific digit (least significant first). */
unsigned g_keys0[PROCS]::[MAX_KEYS_PER_PROC];
unsigned g_keys1[PROCS]::[MAX_KEYS_PER_PROC]; 
unsigned *keys;

/* A list of buckets provides a count of how many keys on a given processor
   have a certain value of the digit being sorted. */
unsigned g_buckets[PROCS]::[RADIX];
unsigned *buckets;

/* A list of global offsets into the key list allows each processor to
   write its keys into their new locations. */
int g_offset[PROCS]::[RADIX];
int *offset;

int g_scan_result[PROCS]::[RADIX];
int *scan_result;

int to_self;   /* count the number of keys I send to myself */

/* all_create_keys fills in the key list with random 32-bit numbers */
/* distribution specifies the type */
void all_create_keys(int dist)
{
  int i;  
	
  srand(MYPROC*17);      
    if (dist) {
      for (i = 0; i < keys_per_proc; i++) 
	keys[i] = rand() & rand() & rand() & rand() ; 
    }
    else {
      for (i = 0; i < keys_per_proc; i++) 
	keys[i] = rand(); 
    }
  barrier();
}

/* all_fill_buckets counts the number of keys with each digit on each
   processor. */
void all_fill_buckets(int pass)
{
	int i, shift = pass*BITS_PER_DIGIT;
	
	for (i = 0; i < RADIX; i++)
		buckets[i] = 0;
	
	for (i = 0; i < keys_per_proc; i++)
		buckets[(keys[i]>>shift)&(RADIX - 1)]++;
}

/* all_scan_buckets uses the buckets histograms produced on each processor
   to determine the global key list offset used by each processor for each
   digit. */
void all_scan_buckets()
{
  int i;

  for (i = 0; i < RADIX; i++) {
    if (MYPROC == 0 && i > 0)
      buckets[i] += g_buckets[PROCS - 1][i - 1];
    buckets[i] = all_scan_add(buckets[i]);
	barrier();
  }	
}

void all_wavefront_scan()
{
  int i,b,proc;
  int sum =0;

  for(i=0; i < RADIX ; i++)   /* use -1 as not valid flag */
    offset[i] = -1;

  barrier(); 

  if (MYPROC == 0) {
    for (b = 0; b< RADIX; b++)
      g_offset[1][b] :- buckets[b];
  }
  else {
    if ( MYPROC < (PROCS-1)) {
      for (b = 0; b < RADIX ; b++) {
	while ( offset[b] == -1)  /* spin waiting for store to complete */
	  splitc_poll();           
	g_offset[MYPROC+1][b] :- offset[b]+buckets[b];
      }
    }
    else {                      /* last proc stores to zero for final bcast */
      for (b = 0; b < RADIX ; b++) {
	while ( offset[b] == -1)  /* spin waiting for store to complete */
	  splitc_poll();
	scan_result[b] =offset[b]+buckets[b] ;
	g_scan_result[0][b] :- scan_result[b];
				
      }
    }
  }
  all_store_sync();
  all_bcast_buffer(toglobal(MYPROC,scan_result),sizeof(int)*RADIX);

  if (MYPROC == 0) {
    for (b=0;b<RADIX;b++) {
      buckets[b] = sum+buckets[b];
      sum += scan_result[b];
    }
  }
  else {
    for (b=0;b<RADIX;b++) {
      buckets[b] = sum+buckets[b]+offset[b];
      sum += scan_result[b];
    }
  }
}


/* all_coalesce moves all of a processor's keys to their new locations
   via the offset array. */
void all_coalesce(int pass)
{
  int i, shift = pass*BITS_PER_DIGIT;
  int key_bucket, new_row, new_column;
	
  if (pass&1) {
    for (i = keys_per_proc - 1; i >= 0; i--) {
#ifdef SC_PROFILE
      prof_compute();
#endif
      key_bucket = (keys[i]>>shift)&(RADIX - 1); 
      buckets[key_bucket]--;
      new_row = (buckets[key_bucket]>>log_keys);
      new_column = buckets[key_bucket]&(keys_per_proc - 1);
      g_keys0[new_row][new_column] :- keys[i];
      if (new_row == MYPROC) 
	to_self++;
    }
  }
  else {
    for (i = keys_per_proc - 1; i >= 0; i--) {
#ifdef SC_PROFILE
      prof_compute();
#endif
      key_bucket = (keys[i]>>shift)&(RADIX - 1); 
      buckets[key_bucket]--;
      new_row = (buckets[key_bucket]>>log_keys);
      new_column = buckets[key_bucket]&(keys_per_proc - 1);
      g_keys1[new_row][new_column] :- keys[i];
      if (new_row == MYPROC)
	to_self++;
    }
  }
  all_store_sync();
}

void splitc_main(int argc, char *argv[])
{
  int pass, i, j;

  /* timing data */
  double time_start, time_finish;
  double total_fill, total_scan, total_coallesce;

  double pass_time[DIGITS_PER_KEY][4];
  int errors;
  int procs;
  int which_scan =0;  /* type of scan  -- a tree or wave */
  int key_dist = 0;   /* key distribution random or 4 times anded */

  keys_per_proc = MAX_KEYS_PER_PROC;
  log_keys = LOG_KEYS;
  procs = PROCS;

  /* parse command line args */
  argv++;
  argc--;
  while (argc>0) {
    if (**argv=='-')
      switch (*((*argv)+1)) {
      case 'h':
	if (MYPROC == 0) {
	  printf("Radix Sort options: (defaults)\n");
	  printf("  -n number of keys (in K) per processor  (512K)\n");
#ifdef CM5	  
	  printf("  -p number of procesors\n");
#endif	  
          printf("  -s do wavefront scan\n");
	  printf("  -e entropy the keys \n");
	}
	exit(0);
	break;
      case 'n':
	if (--argc) sscanf(*++argv, "%d", &keys_per_proc);
	keys_per_proc*= 1024;
	log_keys =0;
	for (i=1; i!=keys_per_proc; i=i<<1)
	  log_keys++;
	break;
#ifdef CM5	
      case 'p':
	if (--argc) sscanf(*++argv, "%d", &procs);
	if ((procs > 0) &&(procs <= PROCS))
	  all_set_PROCS(procs);
	break;
#endif	
      case 's':
	which_scan =1;
	break;
      case 'e':
	key_dist =1;
	break;
      default:
	on_one {
	  printf("unknown option %c\n",(*((*argv)+1)) );
	}
	break;
      }
    else printf("Argument %s meaningless.\n",*argv);
    argc--;
    argv++;
  }	
  
  /* Announce start of sort. */
  on_one {
    printf ("Input size: %d 32-bit numbers\n", keys_per_proc*PROCS);
    printf ("Radix: %d (%d bits)\n\n", RADIX, BITS_PER_DIGIT);
    fflush(stdout);
  }
  
  /* Set up local pointers to global arrays. */
  keys = (unsigned *)&g_keys0[MYPROC][0];
  buckets = (unsigned *)&g_buckets[MYPROC][0];
  offset = (int *)&g_offset[MYPROC][0];
  scan_result = (int *) &g_scan_result[MYPROC][0];
  
  /* Make up an input. */
  all_create_keys(key_dist);
  
  /* Sort the numbers. */
  time_start = get_seconds();
  for (pass = 0; pass < DIGITS_PER_KEY; pass++) {
    pass_time[pass][0] = get_seconds();
    
    keys = (unsigned *) (pass&1 ? &g_keys1[MYPROC][0] : &g_keys0[MYPROC][0]);
    /* Count the number of keys with each value of a digit. */
    all_fill_buckets(pass);
    
    pass_time[pass][1] = get_seconds();
    /* Use the bucket histogram to find global offsets. */
    if (which_scan) 
      all_wavefront_scan();
    else 
      all_scan_buckets();

    barrier();
    reset_store_sync();
    pass_time[pass][2] =  get_seconds();
    /* Move the keys to their new positions. */
    all_coalesce(pass);
    pass_time[pass][3] = get_seconds(); 

  }
  time_finish = get_seconds();

  /* Check that the sort is correct. */
  keys = (unsigned *)
    (pass&1 ? &g_keys1[MYPROC][0] : &g_keys0[MYPROC][0]);
  for (i = 0, errors = 0; i < keys_per_proc - 1; i++)
    if (keys[i] > keys[i + 1])
      errors++;
  if (MYPROC < PROCS - 1) {
    if (pass&1) {
      if (keys[i] > g_keys1[MYPROC + 1][0])
	errors++;
    }
    else {
      if (keys[i] > g_keys0[MYPROC + 1][0])
	errors++;
    }
  }
  errors = all_reduce_to_one_add(errors);
  on_one {
    printf ("%d errors found in sort.\n",errors);
  }
  
  /* Print results. */
  on_one {
    printf ("Sort complete.\n\n");
    total_fill = total_scan = total_coallesce = 0;
    for (i = 0; i < DIGITS_PER_KEY; i++) {
      for (j = 3; j > 0; j--)
	pass_time[i][j] -= pass_time[i][j - 1];
      printf ("Pass %d: ",i);
      if (which_scan) 
	printf ("Fill: %.3f sec,  Wave-scan: %.3f sec,  Coallesce: %.3f sec\n",
		pass_time[i][1], pass_time[i][2], pass_time[i][3]);
      else 
	printf ("Fill: %.3f sec,  Tree-scan: %.3f sec,  Coallesce: %.3f sec\n",
		pass_time[i][1], pass_time[i][2], pass_time[i][3]);

      total_fill += pass_time[i][1];
      total_scan += pass_time[i][2];
      total_coallesce += pass_time[i][3];
    }


    /* lame print all on-one line hack */
    i = 0;

      if (which_scan) 
	printf ("XXX Procs %d keys/proc %d passes %d key_dist %d Fill: %.3f sec,  Wave-scan: %.3f sec,  Coallesce: %.3f sec Total %.3f\n", 
		PROCS, keys_per_proc, DIGITS_PER_KEY,key_dist,
		total_fill,total_scan,total_coallesce,		
		time_finish - time_start);
      else 
	printf ("XXX Procs %d keys/proc %d passes %d key_dist %d Fill: %.3f sec,  Tree-scan: %.3f sec,  Coallesce: %.3f sec Total %.3f\n", 
		PROCS, keys_per_proc, DIGITS_PER_KEY,key_dist,
		total_fill,total_scan,total_coallesce,		
		time_finish - time_start);
    
    printf ("\nTotal Time: %.3f\n\n", time_finish - time_start);
    printf ("MSorts per second: %.3f\n", (keys_per_proc*PROCS)/
	    ((time_finish - time_start)*1000000));
    printf ("KSorts per second per processor: %.3f\n",
	    keys_per_proc/((time_finish - time_start)*1000));
  }

}

