/* 
 * sample.sc - Split-C code for sample sort
 */
#include <stdio.h>
#include <split-c/split-c.h>
#include <split-c/com.h>
#include <split-c/atomic.h>

#define PROCSMAX 64

/*
 * Space for the samples from which the splitters are selected
 */
#define MAX_OVERSAMPLING    512
int     OVERSAMPLING = MAX_OVERSAMPLING;
int g_sample1[PROCS]::[PROCSMAX*MAX_OVERSAMPLING];
int g_sample2[PROCS]::[PROCSMAX*MAX_OVERSAMPLING];


/**
 ** How detailed the timing should be
 **/
int detailed_timing = 0;

/**
 ** How detailed the timing should be
 **/
int verbose = 0;


/**
 ** Should I do presorting for better splitters
 **/
int DO_PRESORT = 1;

/**
 ** Maximum expansion allowed
 **/
double EXPANSION_FACTOR  = 1.4;

/* 
 * Splitter array -- identical on each processor
 */ 
int g_splitter[PROCS]::[PROCSMAX+1]; 
int *splitter;

/*
 * The destination processor of keys matching the splitters 
 */
int duplicate[PROCSMAX];

/*
 * The number of keys sent to this processor
 */
int my_num_keys;

/*
 * The number of keys worth of space allocated on each processor       
 */
#define EXPANSION(x)    (int)(EXPANSION_FACTOR*x)
int bucket_size;

/**
 ** buffers, counts, starting positions
 ** for transfer from processor to processor and
 **/
int *my_transfer_buffs[PROCSMAX];
int my_transfer_count[PROCS]::[PROCSMAX];
int *you_transfer_here[PROCS]::[PROCSMAX];


/*
 * The time of the entire algorithm and the individual phases
 */
#define MAX_PHASES 40
double phase_time[MAX_PHASES];
double time_finish, time_start;
double time_compute, time_comm;

double last_time, cpu_time, net_time;
#define START_TIME()	{last_time = get_seconds();}
#define SUM_TIME(sum)	{double t=get_seconds(); sum+=t-last_time; last_time=t;}

/**
 ** Names of phases. Number of phases.
 **/
char * phase_name[MAX_PHASES];
int num_phases = 0;

/**
 **
 **/
#define NEW_PHASE(name) {phase_time[num_phases] = get_seconds(); phase_name[num_phases++] = name;}




/**
 ** Procedure: print_results
 ** Parameter:
 **   np (input): the number of keys per processor
 **/
void print_results(int np)
{
  int max_keys, i;
  double expansion_factor;

  max_keys = all_reduce_to_one_max(my_num_keys);
  on_one {
    expansion_factor = (double)max_keys / (double)np;
    printf("\nProcs: %d, n/p: %d, expansion %.3f\n", PROCS, np,
           expansion_factor);
#if 0
    for (i = 0; i < num_phases - 1; i++)
      printf("   Phase %2d-%2d: %.4lf   (%s)\n", i, i+1,
	 phase_time[i+1]-phase_time[i], phase_name[i + 1]);
#endif
    printf("Total_Time: %.3fs %.3fus/key/proc, \n",
	time_finish-time_start, (time_finish-time_start)*1E6/PROCS/np);
    printf("@cmd sample-ucsb -n %d\n", np/1024);
    printf("@cpu %.3f\n@net %.3f\n@total %.3f\n",
	cpu_time, net_time, cpu_time+net_time);
  }
}

/*
 * Procedure: duplicate_splitters
 * Parameter:
 *   splitter (input): The list of P splitters
 *
 * duplicate[i] is set to the processor that this processor should
 * send a key that matches splitter[i]; every processor has a
 * different duplicate array so that if there are a lot of these
 * duplicate keys, then they are sent to different processors. 
 */
void duplicate_splitters()
{
  int p, i, a, first;
  
  /*
   * If there are not a sequence of identical splitters, then
   * if a key matches a duplicate, the even processors send the
   * key to the lower processor; the odd processors to the higher
   * processor.
   */
  for (p = 1; p < PROCS; p++) 
    duplicate[p] = p-1 + MYPROC%2;

  /*
   * Look for identical splitters.
   */
  a = splitter[1];
  for (p = 1; p < PROCS; ) {
    first = p;
    while (splitter[p] == a) p++;
    /*
     * There are (p-first+1) identical splitters.
     * Divide the processors into (p-first+1) groups and
     * have each group send to a different one of the (p-first+1) processors
     */
    for (i = first; i < p; i++)
      duplicate[i] = first-1 + MYPROC%(p - first + 1);
    a = splitter[p];
    p++;
  }
}


/*
 * Procedure: sel_bucket
 * Parameters:
 *   v (input): the key
 *   splitter (input): the list of splitters which determine
 *                     the destination processor of this key
 *
 * Returns: the destination processor of this key.
 */
static inline int sel_bucket(int v)
{
  int m, l=0, h=PROCS;
  int tmp;

  while(l+1 < h) {
    m = (l+h)>>1;
    if (v > (tmp = splitter[m])) l = m;
    else if (v < tmp) h = m;
    else return duplicate[m];
  }
      
  return l;
}

/*
 * Procedure: push_atomic1
 * Parameters:
 *   p (input): the destination buffer on this processor
 *   key1 (input): the key to save
 *
 * The destination buffer acts as a stack; my_num_keys is
 * the number of keys in the stack.  We have only allocated
 * space for bucket_size keys, so we must check that we don't overflow
 * our buffer space.
 */
void push_atomic1(int *p, int key1)
{
  p[my_num_keys] = key1;
  if (my_num_keys < bucket_size) my_num_keys++;
}


/**
 ** Procedure: compute_transfer_params
 ** Parameters:
 **   np (input): number of keys
 **   a (input):  array of my keys
 **/
void compute_transfer_params(int np, int *in, int *out)
{
  int *l_my_t_count = tolocal(my_transfer_count);
  int i, j, b, tmp, offset;
  int offs[PROCSMAX];
  int s, L, R, m, bm, bm1;
  int t_offs[PROCSMAX], t_myc[PROCSMAX];


  if(DO_PRESORT == 0)
    for(i = 0; i < np; i++)
      out[i] = in[i];

  if (DO_PRESORT == 0) {
    for(i = 0; i < PROCS; i++)
      l_my_t_count[i] = 0;

    for(i = 0; i < np; i++){
       b = sel_bucket(out[i]);
      (l_my_t_count[b])++;
    }
  }

  if (DO_PRESORT == 1){
	t_offs[0] = 0;
	for(s = 0 ; s < PROCS - 1; s++){
	  L = -1; R = np;
	  while (L < R){
	    m = (L + R + 2) >>1;
	    m--;
	    if (m >= 0)
	      bm = sel_bucket(out[m]);
	    else
	      bm = -1;
	    if (m < np - 1)
	      bm1 = sel_bucket(out[m + 1]);
	    else
	      bm1 = PROCS;

            if (bm > s)
	      R = m;
	    else if (bm < s){
	     if (bm1 > s)
	       L = R = m; /* 0 elements to this proc */
	     else
	       L = m + 1;
            }
	    else if (bm1 == s) /* bm == s here */
	      L = m + 1;
	    else /* bm1 > s */
	     R = L = m;
	   }

 	  if (R != L)
	    fprintf(stderr, "%d: err L != R\n", MYPROC);
	  if (R != m)
	    fprintf(stderr, "%d: err R != m\n", MYPROC);
	  if (m >= 0)
	     bm = sel_bucket(out[m]);
	  else
	    bm = 0;
          if (m < np - 1)
	    bm1 = sel_bucket(out[m + 1]);
	  else{
	    bm1 = PROCS - 1;

	    fprintf(stderr, "AAAAA\n");
	  }
	  if ((bm > s) || ( bm1 <= s))
	    fprintf(stderr, "%d: err 2 ( m = %d [%d, %d])\n", 
				MYPROC, m, bm, bm1);
	  t_offs[s + 1] = m + 1;
	  t_myc[s] = t_offs[s + 1] - t_offs[s];  
	}
	t_myc[PROCS - 1] = np - t_offs[PROCS - 1];
  }
  
  if (DO_PRESORT == 1) {
    for(i = 0; i < PROCS; i++)
     l_my_t_count[i] =t_myc[i];
  }

  offset = 0;
  for(i = 0; i < PROCS; i++){
    offs[i] = offset;
    offset += l_my_t_count[i];
  }

  for(i = 0; i < PROCS; i++)
    my_transfer_buffs[i] = &(in[offs[i]]);

  if (DO_PRESORT == 0){
    for(i = 0; i < PROCS; i++)
      l_my_t_count[i] = 0;

    for(i = 0; i < np; i++){
      b = sel_bucket(tmp = out[i]);
      my_transfer_buffs[b][l_my_t_count[b]] = tmp; 
      (l_my_t_count[b])++;
    }
  }
}
  
/**
 ** Compute your_transfer_here
 **/
void compute_your_transfer(int *dest_array, int np)
{
  int proc, offset;
  int **l_you_t_here = tolocal(you_transfer_here);

#ifdef ALL_OLD_BARRIERS
	barrier();
#endif


  for(proc = 0, offset = 0; proc < PROCS; proc++){
    if (offset > EXPANSION(np)){
	fprintf(stderr, "proc %d: Expansion too big (more than %d)\n",
			 MYPROC, offset);
	exit(1);
    }
    l_you_t_here[proc] = &(dest_array[offset]);
    offset += my_transfer_count[proc][MYPROC];
  }
  if (offset > EXPANSION(np)){
    fprintf(stderr, "proc %d: Expansion too big (more than %d)\n",
		 MYPROC, offset);
    exit(1);
  }
  my_num_keys = offset;

#ifdef ALL_OLD_BARRIERS
	barrier();
#endif
}

/**
 ** Procedure: store_block
 ** Parameters:
 **   proc_n (input): destination processor
 **   my_transfer_count:
 **   you_transfer_here:
 **/
void store_block(int proc_n)
{
  int block = 4*1024;

  int *source_lptr = my_transfer_buffs[proc_n];
  int *global dest_gptr =
    toglobal(proc_n, you_transfer_here[proc_n][MYPROC]);
  int count = my_transfer_count[MYPROC][proc_n];
  int i, size, tmp;


  if (count > 0)
    bulk_store(dest_gptr, source_lptr, sizeof(int)*count);
  /*
  for(i = 0; i < count; i += block){
    size = block; 
    if ((tmp = (count - i)) < size)
	size = tmp;
    if (size > 0)
      bulk_store(dest_gptr, source_lptr, sizeof(int)*size);
    source_lptr += block;
    dest_gptr += block;
  }
  */
}

/**
 ** Procedure: my_all_to_all_store_blocks
 ** Parameters:
 **   my_transfer_count:
 **   you_transfer_here:
 **/
void my_all_to_all_store_blocks()
{
  int i;
  int dest_proc;

  for(i = 1; i < PROCS; i++){
    dest_proc = (MYPROC + i) % PROCS;
    store_block(dest_proc);
    barrier(); /*all_store_sync(); */
  }
  
  /**
   ** Transfer my own elements. Doing it the lazy way.
   ** Optimize later.
   **/
  store_block(MYPROC);
  all_store_sync();
}

/**
 ** Procedure: my_all_to_all_store_blocks
 ** Parameters:
 **   my_transfer_count:
 **   you_transfer_here:
 **/
void old_my_all_to_all_store_blocks()
{
  int i, n_phases, phase, group_size, half_gsize;
  int pard_halfgroup, pard_halfgroup_offs, pard_i, pard_num;
  int my_halfgroup, my_i, my_half;

  for(i = PROCS, n_phases = 0; (i = i >> 1) != 0; n_phases++)
    ;

  for(phase = 0, group_size = PROCS; phase < n_phases;
      phase++, group_size = group_size >> 1)
  {

    half_gsize = group_size >> 1;
    my_halfgroup = (MYPROC / half_gsize);
    my_i = MYPROC % half_gsize;
    my_half = my_halfgroup % 2;
    pard_halfgroup =  my_half?(my_halfgroup - 1):(my_halfgroup + 1);
    pard_halfgroup_offs = half_gsize*pard_halfgroup;
    for(i = 0; i < half_gsize; i++)
    {
	pard_i = (my_i + i) % half_gsize;
	pard_num = pard_halfgroup_offs + pard_i;
	store_block(pard_num);
        all_store_sync();
    }

  }

  /**
   ** Transfer my own elements. Doing it the lazy way.
   ** Optimize later.
   **/
  store_block(MYPROC);
  all_store_sync();

/*  barrier(); */
}


void sample_sort(unsigned int np, int *in, int *out)
{
  static int sample[MAX_OVERSAMPLING];
  int *all_samples1, *all_samples2, *global sample_ptr;
  int *nextKey,*addr, i, j;
  int p;

  START_TIME();

  if (DO_PRESORT) {
    /**
     ** Presort in order to get  better splittrs
     **/

#ifdef ALL_OLD_BARRIERS
	barrier();
#endif

    local_radix_sort(np, in, out);
    for(i = 0; i < np; i++)
	in[i] = out[i];

#ifdef ALL_OLD_BARRIERS
	barrier();
#endif

  }

  /* 
   * Every processor selects every np/OVERSAMPLE-th key
   * as its sample set; processor 0 puts these keys directly
   * in the all_samples buffer; the other processor put the
   * samples in a tempory buffer
   */

  if (MYPROC == 0) {
    all_samples1 = (int *)g_sample1;
    all_samples2 = (int *)g_sample2;
    for(i = 0; i < OVERSAMPLING; i++) 
      all_samples1[i] = in[i * (np / OVERSAMPLING)];
  }
  else {
    addr = (int *)g_sample1;
    addr += MYPROC*OVERSAMPLING;
    sample_ptr = toglobal(0, addr);

    for(i = 0; i < OVERSAMPLING; i++) 
      sample[i] = in[i * (np / OVERSAMPLING)];
  }
  barrier();
  SUM_TIME(cpu_time);

  /*
   * All processors now send the samples to processor 0 
   * The problem is, CMAM has a limit of 256 segments that can
   * be open at one time; therefore, we only let 256 processors
   * send at once.
   */
  for (i = 0; i < PROCS; i += 256) {
    if ((i < MYPROC)  && (MYPROC < (i+256))) 
      bulk_store(sample_ptr, sample, sizeof(sample));
    all_store_sync();
  }
  SUM_TIME(net_time);

  /* 
   * Sort samples on processor 0, select every OVERSAMPLING-th sample as
   * the splitters, and broadcast the all the splitters to all processors.
   */

  if (MYPROC == 0) {
    local_radix_sort_8bit(PROCS*OVERSAMPLING, all_samples1, all_samples2);
    splitter[0] = 0x80000000;   
    for(i = 1; i < PROCS; i++)
      splitter[i] = all_samples1[i*OVERSAMPLING];
  }
  my_num_keys=0;
  barrier();
  SUM_TIME(cpu_time);

  all_bcast_buffer(toglobal(MYPROC, splitter), sizeof(int)*PROCS);
  splitter[PROCS] = 0x7fffffff;

#ifdef SOME_NEW_BARRIERS
  barrier();
#endif

  SUM_TIME(net_time);

  /*
   * Determine the destination processor for those keys that match
   * one of the splitters.
   */
  duplicate_splitters();

  /*
   * We look at each of our keys, determine its destination processor,
   * and push the key onto the destination processor's local stack.
   */
  compute_transfer_params(np, in, out);
  compute_your_transfer(out, np);

/* 
 * GC: compute_your_transfer() ends with a barrier() if ALL_OLD_BARRIERS set
 */  
#ifndef ALL_OLD_BARRIERS
#ifdef  SOME_NEW_BARRIERS
	barrier();
#endif
#endif

  SUM_TIME(cpu_time);

  my_all_to_all_store_blocks();
  SUM_TIME(net_time);

  /*
   * After we have received all of our keys, perform a local sort on
   * them; keys end up in "in" buffer.
   */
  local_radix_sort(my_num_keys, out, in);
  barrier();
  SUM_TIME(cpu_time);
}                             

void splitc_main(int argc, char**argv)
{
  extern char *optarg;
  int i, j, dist=0, *in, *out, np = 524288;
  signed char c;
  double save_time_start;
  int choose_expansion = 1;
  int choose_presort = 1;

  /* Get options from the user */
  while ((c = getopt(argc, argv, "P:n:d:o:x:v:p:t:h")) != -1) {
    switch(c) {
    case 'n':  	np = atoi(optarg);  np *= 1024;
	break;
    case 'd':  	dist = atoi(optarg);
	break;
    case 'o':  	OVERSAMPLING = atoi(optarg);
	break;
    case 'x':  	EXPANSION_FACTOR = atoi(optarg) / 10.0;
		choose_expansion = 0;
	break;
    case 'p':  	DO_PRESORT = atoi(optarg);
		 choose_presort = 0;
	break;
    case 'v':  	verbose = atoi(optarg);
	break;
    case 't':  	detailed_timing = atoi(optarg);
	break;
#ifdef CM5
    case 'P':  	all_set_PROCS(atoi(optarg));
	break;
#endif
    case 'h':
      on_one {
        printf("Sample sort options: (defaults)\n");
        printf("  -n number of keys (in K) per processor  (64K)\n");
        printf("  -d key distribution (0 - 4)\n");
        printf("  -o oversampling (64 - 512\n");
        printf("  -x expansion factor (10 - 30)\n");
        printf("  -p presort (0 or 1)\n");
        printf("  -v verbose (0 or 1)\n");
        printf("  -P PROCS\n");
      }
      break;
    }
  }
  on_one printf("Sorting %d ints (%d ints/proc)\n", PROCS*np, np);


/****
  on_one printf("PROCSMAX = %d\n", PROCSMAX);
****/

  if (dist == 0){
    if (choose_presort)
      DO_PRESORT = 0;
    if (choose_expansion) 
      EXPANSION_FACTOR = 1.25;
  }

  if (dist > 0){
    if (choose_presort)
      DO_PRESORT = 1;
    if (choose_expansion) 
      EXPANSION_FACTOR = 1.45;
  }

  barrier();
  save_time_start = time_start = get_seconds();  
  /* Allocate and initialize arrays */
  splitter = (int *)g_splitter;
  bucket_size = EXPANSION(np);
 
/*
  in = (int *)all_spread_malloc(PROCS, sizeof(int) * bucket_size);
  out = (int *)all_spread_malloc(PROCS, sizeof(int) * bucket_size);
*/
  in = (int *)malloc(sizeof(int) * bucket_size );
  out = (int *)malloc(sizeof(int) * bucket_size );


  if ((in == NULL) || (out == NULL)) {
    fprintf(stderr, "proc %d: all_spread_malloc failed! exiting.\n",MYPROC);
    exit(-1);
  }
  
#if 0
   time_finish = get_seconds();
   printf("  %d: Alloc took %f sec\n", MYPROC, time_finish - time_start);
   time_start = get_seconds(); 
#endif

   create_keys(np, in, dist);

#if 0
  time_finish = get_seconds();
   printf("        %d: Create took %f sec\n", MYPROC, time_finish - time_start);
   time_start = get_seconds();  
#endif
 
  for(i = 0; i < bucket_size; i++)
    out[i]  = 0;

#if 0
  tie_finish = get_seconds();
   printf("              %d: Clear took %f sec\n", MYPROC, time_finish - time_start);
#endif

   barrier();
  time_finish = get_seconds();
  time_start = save_time_start;
  on_one {
    printf(" Setup took %f sec\n", time_finish - time_start);
    fflush(stdout);
  }

  barrier();
  time_start = get_seconds();

  sample_sort(np, in, out);
  if (verbose) {
    printf("%d: done sorting\n",MYPROC);
    fflush(stdout);
  }
  barrier();
  time_finish = get_seconds();
  
  all_check_results(my_num_keys, in, 0);

  print_results(np);

  printf("%d: keys = %d\n", MYPROC, my_num_keys);
  fflush(stdout);

  my_num_keys = all_reduce_to_one_add(my_num_keys);
  on_one {
    printf("%d: Allkeys = %d\n", MYPROC, my_num_keys);
    fflush(stdout);
  }

 barrier();
}

