/*									tab:4
 *
 * sample.sc - sample sort
 *
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: 			Richard Martin
 * Version:			2
 * Creation Date:	Mon Aug 16 17:59:19 1993
 * Filename:		sample.sc
 * History:
 *	LTL	1	Mon Aug 16 17:59:54 1993
 *		Added splitc_poll() to speed up CMMD version
 */



/* a sample sort using only split-c 
 * based on the sample sort created with Thortsen Von Eiken
 * note that this version does not tag keys with proc number to
 * break ties
 */

#include <stdio.h>
#include <split-c/split-c.h>
#include <split-c/com.h>
#include <split-c/atomic.h>
const int terse = 1;

#define	OVERSAMPLING	16
#define DPROCS          64  
#define DEFAULT_KEYS    262144 /* number of keys */
#define EXPANSION(x)    ((2*x)/1)

int *global sample_ptr[PROCS]:: ; /* points to where sample goes on proc 0*/
int splitter[PROCS]::[DPROCS]; /* splitter array*/

/* figure out how big each buckets if on the processors */
int check[PROCS]::;
int n,num_keys,bucket_size;

double phase_time[6];

static inline int
sel_bucket(int v, int *splitter)
{
	int m, l=0, h=PROCS;
	while(l+1 < h) {
		m = (l+h)>>1;
		if(v >= splitter[m]) l = m; else h = m;
	}
	if(v < splitter[l] || (v >= splitter[l+1] && !(l==PROCS-1)))
		printf("sel_buckets error: not %d <= %d < %d (l=%d)\n",
			splitter[l], v, splitter[l+1], l);
	return l;
}

void update_atomic2(int *p,int key1, int key2)
{
    p[num_keys]=key1;
    p[num_keys+1]=key2;
    num_keys+=2;
    if (num_keys > bucket_size) {
      printf("proc %d: Too many keys, dropping some\n",MYPROC);
      num_keys-=2;
    }
}

void update_atomic1(int *p,int key1)
{
  p[num_keys]=key1;
  num_keys++;
  if (num_keys > bucket_size) {
    printf("proc %d: Too many keys, dropping some\n",MYPROC);
    num_keys--;
  }
}

void sample_sort(unsigned int n, int *in, int *out)
{
  static int sample[OVERSAMPLING];
  int *lsplitter = tolocal(splitter);
  int offset,errors;
  int *all_samples1, *all_samples2;
  int *nextKey;
  int packet_size[PROCS];
  int packet_data[PROCS];
  int i,j;
  register int p;

  /* all form sample to send to proc 0*/
  phase_time[0] = get_seconds();
  for(i=0; i<OVERSAMPLING; i++) {
    sample[i] = in[i*(n/OVERSAMPLING)];
  }
  /* collect samples on processor 0 */
  if(MYPROC == 0) {
    all_samples1 = (int *)malloc(PROCS*sizeof(sample));
    all_samples2 = (int *)malloc(PROCS*sizeof(sample));
    for (i=1;i<PROCS;i++) 
      sample_ptr[i] := toglobal(0,&all_samples1[i*OVERSAMPLING]);
    
    for (i=0;i<OVERSAMPLING;i++)
      all_samples1[i] = sample[i];
    sync();
  }

  barrier();
  /* now all send samples to proc 0 */
  if (MYPROC != 0) {
    bulk_put(sample_ptr[MYPROC],sample,sizeof(sample));
    sync();
  }
  barrier();
  phase_time[1] = get_seconds();

  /* sort samples on processor 0 and broadcast splitters */
  if(MYPROC == 0) {
    if(!terse) printf("Sorting samples.\n");
    local_radix_sort(PROCS*OVERSAMPLING,all_samples1,all_samples2);
    if(!terse) {
      printf("Sorted samples:\n");
      for(i=0; i<PROCS*OVERSAMPLING; i++)
	printf("%7d ", all_samples2[i]);
      printf("\n");
    }
    lsplitter[0] = 0x80000000;	
    for(i=1; i<PROCS; i++)
      lsplitter[i] = all_samples1[i*OVERSAMPLING];
    free(all_samples1); free(all_samples2);

    if(!terse) printf("Broadcasting splitters.\n");
    if(!terse) {
      for(i=0; i<PROCS; i++) printf("%x ", lsplitter[i]);
      printf("\n");
    }
    for(i=1; i<PROCS; i++) 
      bulk_put(splitter[i],lsplitter,sizeof(int)*PROCS);
    sync();
  }

  num_keys=0;
  barrier();
  for (i=0;i<PROCS;i++) {
   packet_size[i] = packet_data[i] = 0;

  }
  phase_time[2] = get_seconds();
  /* now all procs do not build a histogram, but store the keys 
   * with atomic procedures 
   * this version (003) stores keys in packets, sends 2 keys at once 
   */
  if(!terse && MYPROC == 0) printf("Storing keys.\n");
  for(i=0; i<n; i++) {
    p = sel_bucket(in[i],lsplitter);
    if (packet_size[p]) {
      atomic(update_atomic2,toglobal(p,out),in[i],packet_data[p]);
      packet_size[p]=0;
    }
    else {
      packet_data[p] = in[i];
      packet_size[p]++;
    }
    splitc_poll();
  }

  for (i=0;i<PROCS;i++) {   /* now clean up all leftover packets */
    if (packet_size[p]) 
      atomic(update_atomic1,toglobal(p,out),packet_data[p]);
    splitc_poll();
  }

  all_atomic_sync();
  phase_time[3] = get_seconds();

  /* now do a local radix sort */
  if(!terse && MYPROC == 0) printf("Local Sort of %d keys.\n",num_keys);
  local_radix_sort(num_keys,out,in);
}							  
void
init_array(unsigned int n, int data[n])
{
  int i;
  srand(MYPROC*17);
  for(i=0; i<n; i++)
  {
	  data[i] = (int) rand() & 0xffff;
	  data[i] |= (int) rand() << 16;
  }
}

void
splitc_main(int argc, char**argv)
{
	static int n_array[PROCS]::;
	double time_start,time_finish;
	int i,errors;
	int *in;
	int *out;

	n = DEFAULT_KEYS;

	argv++;
	argc--;
	while (argc>0) {
	  if (**argv=='-')
	    switch (*((*argv)+1)) {
	    case 'h':
	      if (MYPROC == 0) {
		printf("Sample Sort options: (defaults)\n");
		printf("  -n number of keys (in K) per processor (512K)\n");
	      }
	      exit(0);
	      break;
	    case 'n':
	      if (--argc) sscanf(*++argv, "%d", &n);
	      n*= 1024;
	      break;
	    }
	  else printf("Argument %s meaningless.\n",*argv);
	  argc--;
	  argv++;
	}	


	/*  broadcast sort size */
	barrier();
	if(MYPROC == 0) {
		printf("Sorting %d ints (%dKints/proc)\n", PROCS*n, n/1024);
		for(i=0; i<PROCS; i++)
			n_array[i] := n;
		sync();
	}
	fflush(stdout);
	barrier();
	n = n_array[MYPROC];

	/* allocate and initialize arrays */
	bucket_size = EXPANSION(n);
	in = (int *)malloc(sizeof(int) * bucket_size);
	out = (int *)malloc(sizeof(int) * bucket_size);
	if ( (in == NULL) || (out == NULL) ) {
	  printf("proc %d: malloc failed! exiting. \n",MYPROC);
	  exit(0);
	}

	init_array(n, in);

	/* do the sort */
	barrier();
	if(!terse && MYPROC == 0) printf("Starting sort.\n");
	time_start= get_seconds();
	sample_sort(n, in, out);
	time_finish=get_seconds();
	if(!terse && MYPROC == 0) printf("Done with sort.\n");
	
	/* Check that the sort is correct. */	
	for (i = 0, errors = 0; i < num_keys-1; i++){
	  if (out[i] > out[i + 1]) {
	    errors++;
	  }
	}
	check[MYPROC]=out[num_keys-1];
	barrier();
	if (MYPROC < (PROCS - 1)) {
	  if (check[MYPROC] > check[MYPROC+1]) {
	    errors++; 
	    printf("proc %d: error across to  %d \n",MYPROC,MYPROC+1);  
	  }
	}

	barrier();

	errors = all_reduce_to_one_add(errors);
	on_one {
	  printf ("%d errors found in sort.\n",errors);
	}
	on_one {
	  printf ("\nProcs: %d Total Time: %.3f Time per key per Proc %.3f n/p: %d\n\n", PROCS, time_finish - time_start, 
		  n/(time_finish - time_start), n);
	}
	on_one {
	  if (1) {
	    printf("time for phases:\n");
	    for (i=0;i<3 ; i++)
	      printf("Phase %d-%d: %lf\n",i+1,i,
		     phase_time[i+1]-phase_time[i]);
	    printf("local_sort %lf\n",time_finish-phase_time[i]);
	  }
	}
}
