/*									tab:4
 *
 * shell.sc - Example program which sorts integers.  This version
 *   uses bulk_reads.
 *
 * Copyright (c) 1993 and The Regents of the University 
 * of California.  All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: 			Andrea C. Dusseau
 * Version:			8
 * Creation Date:	Wed Jan 20 15:19:16 1993
 * Filename:		shell.sc
 * History:
 */
static char _version_string_[] = "\nVersion:1:shell.sc\0\n";

#include <stdio.h>
#include <split-c/split-c.h>
#include <math.h>
#include "shell.h" 

/* -------------------------- splitc_main --------------------- */
/*
 * Procedure: splitc_main
 *
 * Algorithm:  
 * After getting the number of keys to sort per processor and 
 * the blocking factor from the user, each processor generates a
 * random set of integer keys.  These keys are then sorted locally.
 * Processors then merge their keys with a different processor over
 * a series of iterations.  A mop up step is performed to between
 * neighboring processors.  Finally, the list of keys is checked
 * to ensure that it is correctly sorted.
 */ 
void splitc_main(int argc, char *argv[])
{
  int num;
  int b;

  if (argc == 3) {
    num = atoi(argv[1]);
    b = atoi(argv[2]);
  }
  else {
    num = 1000;
    b = 50;
  }
  
  if (num % b) 
    on_one {
      printf("ERROR: Number of keys not evenly divisable by the blocking factor.\n");
      exit(-1);
    }

  on_one  {
    printf("Sorting %d keys per processor with a block transfer size of %d.\n",
	   num, b);
    fflush(stdout);
  }	
  
  {
    int (*spread keys)[num], (*spread merged)[num];
    int d, s;
    double time0, time1;

    if ((keys  = (void *spread)all_spread_malloc(PROCS, num * sizeof(int))) == NULL) {
      on_one {
	printf("ERROR: all_spread_malloc returned NULL.\n");
	exit(-1);
      }
    }
    if ((merged = (void *spread)all_spread_malloc(PROCS, num * sizeof(int))) == NULL){ 
      on_one {
	printf("ERROR: all_spread_malloc returned NULL.\n");
	exit(-1);
      }
    }

    GenRandKeys(tolocal(keys), num); 

    barrier();
    time0 = get_seconds();
    LocalSort(tolocal(keys[MYPROC]), 0, num-1); 
    time1 = get_seconds();
    on_one printf("Local Sort: %fs.\n", time1 - time0);
    barrier();
  
    for (d = MERGE_ITERATIONS; d > 1; d--) { 
      time0 = get_seconds();
      if (MYPROC > MERGE_COLUMN(MYPROC, d))
	MergeGetHigher(num, keys, merged, MERGE_COLUMN(MYPROC,d), b);
      else
	MergeGetLower(num, keys, merged, MERGE_COLUMN(MYPROC,d), b);      
      swap(keys, merged);
      time1 = get_seconds();
      on_one {
	printf("Merge Step: %fs.\n", time1 - time0);
	fflush(stdout);
      }
    }
    
    time0 = get_seconds();
    s = MopUp(num, &keys, &merged, b);
    time1 = get_seconds();
    on_one {
      printf("%d Mop Up steps required: %fs.\n", s, time1 - time0);
      fflush(stdout);
    }
    
    on_one Check(num, keys);
  }
  barrier();
}

/* --------------------------- MopUp --------------------------------- */
int inline MopUp(int num, int (*spread *keys)[num], int (*spread *merged)[num], int b)
{
  int flag, s = 0;

  do {
    if (MYPROC % 2 == 0) 
      MergeGetLower(num, *keys, *merged, MYPROC+1, b);
    else 
      MergeGetHigher(num, *keys, *merged, MYPROC-1, b);
    swap(*keys, *merged);

    /*
     * We have to keep merging if the last element on the previous processor 
     * is greater than our first element
     */
    s++;
    if (MYPROC == 0) flag = 0;
    else flag = ((*keys)[MYPROC - 1][num-1] > (*keys)[MYPROC][0]);
    if (!all_reduce_to_all_add(flag)) return s;

    if ((MYPROC % 2 == 0) && (MYPROC != 0))  
      MergeGetHigher(num, *keys, *merged, MYPROC-1, b);
    else if ((MYPROC != PROCS - 1) && (MYPROC != 0))
      MergeGetLower(num, *keys, *merged, MYPROC+1, b);
    else MergeWithSelf(num, *keys, *merged);
    swap(*keys, *merged);

    s++;
    if (MYPROC == 0) flag = 0;
    else flag = ((*keys)[MYPROC-1][num-1] > (*keys)[MYPROC][0]);
    if (!all_reduce_to_all_add(flag)) return s;

  } while (1);
}

/* -------------------------- MergeGetLower --------------------------- */
/*
 * This processor is the lower one in the pair.  It reads a block of b keys
 * from its partner and puts the lower of those keys in "buffer" and
 * of those in its local portion of "keys" into "merged".  When it
 * reaches the end of the buffer, it reads a new chunk.  After "num" keys
 * have been copies into "merged", "merged" and "keys" are swapped.
 */
inline MergeGetLower(int num, int (*spread keys)[num], int (*spread merged)[num], 
		   int proc, int b)
{
  int buf_i, key_i, merge_i, fetched;
  int buffer[b];

  buf_i = key_i = merge_i = 0;
  fetched = b;

  barrier();
  bulk_read(buffer, keys[proc], b * sizeof(int));
  for (; merge_i < num; merge_i++) {
    /* access the elements in increasing order through the array */
     if (keys[MYPROC][key_i] <= buffer[buf_i]) {
       merged[MYPROC][merge_i] = keys[MYPROC][key_i];
       key_i++;
     }
     else {
       merged[MYPROC][merge_i] = buffer[buf_i];
       if (++buf_i == b) {
	 buf_i = 0;
	 bulk_read(buffer, keys[proc] + fetched, b * sizeof(int));
	 fetched += b;
       }
     }
   }
  barrier();


}

/* -------------------------- MergeGetHigher --------------------------- */
/*
 * This processor is the higher one in the pair.  It reads a block of b keys
 * from its partner and puts the higher of those keys in "buffer" and
 * of those in its local portion of "keys" into "merged".  When it
 * reaches the end of the buffer, it reads a new chunk.  After "num" keys
 * have been copies into "merged", "merged" and "keys" are swapped.
 */
inline MergeGetHigher(int num, int (*spread keys)[num], int (*spread merged)[num], 
		    int proc, int b)
{
  int buf_i, key_i, merge_i, fetched;
  int buffer[b];

  buf_i = b - 1;
  key_i = num - 1;
  fetched = num - b;

  barrier();
  bulk_read(buffer, keys[proc] + fetched, b * sizeof(int));
  for (merge_i = num - 1; merge_i >=0; merge_i--) {
    /* access the elements in decreasing order through the array */
    if (keys[MYPROC][key_i] >= buffer[buf_i]) {
      merged[MYPROC][merge_i] = keys[MYPROC][key_i];
      key_i--;
    }
    else {
      merged[MYPROC][merge_i] = buffer[buf_i];
      if (!buf_i--) {
	buf_i = b - 1;
	fetched -= b;
	bulk_read(buffer, keys[proc] + fetched, b * sizeof(int));
      }
    }
  }
  barrier();

}

/* -------------------------- MergeWithSelf --------------------------- */
/*
 * This is called when a processor does not have a partner.  It
 * simply copies the elements from "keys" into "merged" and then
 * swaps the global pointers.
 */
inline MergeWithSelf(int num, int (*spread keys)[num], int (*spread merged)[num])
{
  int merge_i;

  barrier();
  for (merge_i = 0; merge_i < num; merge_i++)
    merged[MYPROC][merge_i] = keys[MYPROC][merge_i];
  barrier();

}

/* ----------------------------- Check ------------------------------ */
/*
 * This procedure linearly runs through the entire global array of
 * keys, ensuring that each element is greater than or equal to the
 * preceeding element.
 */
inline Check(int num, int (*spread keys)[num])
{
  int i, j;
  int last = 0, this, wrong = 0;

  printf("Checking result. . .\n");
  for (i = 0; i < PROCS; i++)
    for (j = 0; j < num; j++) {
      this = keys[i][j];
      if (this < last) {
	printf("ERROR: %d < %d in sorted list. i==%d, j==%d.\n", this,
	       last, i, j);
	if (wrong++ > 10) {
	  printf("Too many errors!  Exiting!\n");
	  exit(1);
	}
      }
      last = this;
    }
  if (!wrong)  printf("Sorted Correctly!\n");
}

/* -------------------------- GenRandKeys ------------------------------ */
/*
 * Each processor creates a local random list of "num" keys.
 */
void inline GenRandKeys(int keys[], int num)
{     
  int i;

  srandom(MYPROC);
  for(i = 0; i < num; i++)
    keys[i]=(int) random();  
}


/* -------------------------- LocalSort ------------------------------ */
/*
 * The local array of ints keys[first] to keys[last] is sorted in place.
 */
void inline LocalSort(int *keys, int first, int last)
{
  int key;
  int temp;
  int cur_first,cur_last;
  int i;

  if ((last-first)>9)  {
    i = random() % (last - first + 1) + first;
    temp = keys[first];
    keys[first] = keys[i];
    keys[i] = temp;
    cur_first = first + 1;
    cur_last = last;
    key = keys[first];

    while (cur_first < cur_last) {
      while ((keys[cur_first] <= key) && (cur_first <= last)) cur_first++;
      while (keys[cur_last] > key) cur_last--;

      if (cur_first<cur_last) {
	temp = keys[cur_first];
	keys[cur_first] = keys[cur_last];
	keys[cur_last] = temp;
      } 
    }
    LocalSort(keys, first, cur_first-1);
    LocalSort(keys, cur_last+1, last);
  }
  else {
    for (i = first+1; i <= last; i++) {
      cur_first=i;
      while ((cur_first > first) && (keys[cur_first] < keys[cur_first-1])) {
	temp = keys[cur_first];
	keys[cur_first] = keys[cur_first-1];
	keys[cur_first-1] = temp;
	cur_first--;
      }
    }
  }
}
